# Copyright 2019 Rick DeWitt <aa0rd@yahoo.com>
# Version 1.0: CatClone- Implementing fake memory image
# Version 2.0: No Live Mode library links. Implementing mem as Clone Mode
#              Having fun with Dictionaries
# Version 2.1: Adding match_model function to fix File>New issue #7409
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import logging
import threading
from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueInteger, \
    RadioSettingValueFloat, RadioSettings

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0000;
struct {            // 20 bytes per chan
  u32  rxfreq;
  u32  txfreq;
  u8   xmode:4,     // param stored as CAT value
       data:2,
       tmode:2;
  u8   rtone;
  u8   ctone;
  u8   filter:1,
       fmnrw:1,
       skip:1,
       nu:5;
  char name[8];
} ch_mem[120];   // 100 normal + 10 P-type + 10 EXT

struct {        // 5 bytes each
  u32 asfreq;
  u8  asmode:4,     // param stored as CAT value
      asdata:2,
      asnu:2;
} asf[32];

struct {        // 10 x 5 4-byte frequencies
  u32  ssfreq;
} ssf[50];

struct {        // 16 bytes
  u8  txeq;
  u8  rxeq;
} eqx[8];

struct {        // Common to S and SG models
  u8   ag;
  u8   an1:2,
       an2:2,
       an3:2,
       anu:2;
  u32  fa;
  u32  fb;
  char fv[4];
  u8   mf;
  u8   mg;
  u8   pc;
  u8   rg;
  u8   tp;
} settings;

struct {            // Menu A/B settings by TS-590SG names
  char ex001[8];    // 590S values get put in SG equiv
  u8   ex002;       // These params stored as nibbles
  u8   ex003;
  u8   ex005;
  u8   ex006;
  u8   ex007;
  u8   ex008;
  u8   ex009;
  u8   ex010;
  u8   ex011;
  u8   ex012;
  u8   ex013;
  u8   ex016;
  u8   ex017;
  u8   ex018;
  u8   ex019;
  u8   ex021;
  u8   ex022;
  u8   ex023;
  u8   ex024;
  u8   ex025;
  u8   ex026;
  u8   ex054;
  u8   ex055;
  u8   ex076;
  u8   ex077;
  u8   ex087;
  u8   ex088;
  u8   ex089;
  u8   ex090;
  u8   ex091;
  u8   ex092;
  u8   ex093;
  u8   ex094;
  u8   ex095;
  u8   ex096;
  u8   ex097;
  u8   ex098;
  u8   ex099;
} exset[2];

  char mdl_name[9];     // appended model name, first 9 chars

"""

STIMEOUT = 2
LOCK = threading.Lock()
BAUD = 0    # Initial baud rate
MEMSEL = 0  # Default Menu A
BEEPVOL = 4     # Default beep volume
W8S = 0.01      # short wait, secs
W8L = 0.05      # long wait

TS590_DUPLEX = ["", "-", "+"]
TS590_SKIP = ["", "S"]

# start at 0:LSB
TS590_MODES = ["LSB", "USB", "CW", "FM", "AM", "FSK", "CW-R",
               "FSK-R", "Data+LSB", "Data+USB", "Data+FM"]
EX_MODES = ["FSK-R", "CW-R", "Data+LSB", "Data+USB", "Data+FM"]

TS590_TONES = list(chirp_common.TONES)
TS590_TONES.append(1750.0)

RADIO_IDS = {   # From kenwood_live.py; used to report wrong radio
    "ID019;": "TS-2000",
    "ID009;": "TS-850",
    "ID020;": "TS-480",
    "ID021;": "TS-590S",
    "ID023;": "TS-590SG"
}


def command(ser, cmd, rsplen, w8t=0.01, exts=""):
    """Send @cmd to radio via @ser"""
    # cmd is output string without ; terminator
    # rsplen is expected response char count, including terminator
    #       If rsplen = 0 then do not read after write

    start = time.time()
    #   LOCK.acquire()
    stx = cmd       # preserve cmd for response check
    stx = stx + exts + ";"    # append arguments
    ser.write(stx)
    LOG.debug("PC->RADIO [%s]" % stx)
    ts = time.time()        # implement the wait after command
    while (time.time() - ts) < w8t:
        ix = 0      # NOP
    result = ""
    if rsplen > 0:  # read response
        result = ser.read(rsplen)
        LOG.debug("RADIO->PC [%s]" % result)
        result = result[:-1]        # remove terminator
    #   LOCK.release()
    return result.strip()


def _connect_radio(radio):
    """Determine baud rate and verify radio on-line"""
    global BAUD        # Allows modification
    bauds = [115200, 57600, 38400, 19200, 9600, 4800]
    if BAUD > 0:
        bauds.insert(0, BAUD)       # Make the detected one first
    # Flush the input buffer
    radio.pipe.timeout = 0.005
    radio.pipe.baudrate = 9600
    junk = radio.pipe.read(256)
    radio.pipe.timeout = STIMEOUT

    for bd in bauds:
        radio.pipe.baudrate = bd
        BAUD = bd
        radio.pipe.write(";")
        radio.pipe.write(";")
        resp = radio.pipe.read(4)
        radio.pipe.write("ID;")
        resp = radio.pipe.read(6)

        if resp == radio.ID:           # Good comms
            resp = command(radio.pipe, "AI0", 0, W8L)
            return
        elif resp in RADIO_IDS.keys():
            msg = "Radio reported as model %s, not %s!" % \
                (RADIO_IDS[resp], radio.MODEL)
            raise errors.RadioError(msg)
    raise errors.RadioError("No response from radio")


def read_str(radio, trm=";"):
    """ Read chars until terminator """
    stq = ""
    ctq = ""
    while ctq != trm:
        ctq = radio.pipe.read(1)
        stq += ctq
    LOG.debug("   + [%s]" % stq)
    return stq[:-1]     # Return without trm


def _read_mem(radio):
    """Get the memory map"""
    global BEEPVOL
    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = radio._upper + 20  # 10 P chans and 10 EXT
    status.msg = "Reading Channel Memory..."
    radio.status_fn(status)

    result0 = command(radio.pipe, "EX0050000", 9, W8S)
    result0 += read_str(radio)
    BEEPVOL = int(result0[6:])
    result0 = command(radio.pipe, "EX005000000", 0, W8L)   # Silence beeps

    data = ""
    mrlen = 41      # Expected fixed return string length
    for chn in range(0, (radio._upper + 21)):   # Loop stops at +20
        # Request this mem chn
        r0ch = 999
        r1ch = r0ch
        # return results can come back out of order
        while (r0ch != chn):
            # simplex
            if chn < 100:
                result0 = command(radio.pipe, "MR0 %02i" % chn,
                                  mrlen, W8S)
                result0 += read_str(radio)
            else:
                result0 = command(radio.pipe, "MR0%03i" % chn,
                                  mrlen, W8S)
                result0 += read_str(radio)
            r0ch = int(result0[3:6])
        while (r1ch != chn):
            # split
            if chn < 100:
                result1 = command(radio.pipe, "MR1 %02i" % chn,
                                  mrlen, W8S)
                result1 += read_str(radio)
            else:
                result1 = command(radio.pipe, "MR1%03i" % chn,
                                  mrlen, W8S)
                result1 += read_str(radio)
            r1ch = int(result1[3:6])
        data += radio._parse_mem_spec(result0, result1)
        # UI Update
        status.cur = chn
        status.msg = "Reading Channel Memory..."
        radio.status_fn(status)

    if len(data) == 0:       # To satisfy run_tests
        raise errors.RadioError('No data received.')
    return data


def _make_dat(sx, nb):
    """ Split the string sx into nb binary bytes """
    vx = int(sx)
    dx = ""
    if nb > 3:
        dx += chr((vx >> 24) & 0xFF)
    if nb > 2:
        dx += chr((vx >> 16) & 0xFF)
    if nb > 1:
        dx += chr((vx >> 8) & 0xFF)
    dx += chr(vx & 0xFF)
    return dx


def _sets_val(stx, nv=3, nb=2):
    """ Split string stx into nv nb-bit values in 1 byte """
    # Right now: hardcoded for nv:3 values of nb:2 bits each
    v1 = int(stx[0]) << 6
    v1 = v1 | (int(stx[1]) << 4)
    v1 = v1 | (int(stx[2]) << 2)
    return chr(v1)


def _sets_asf(stx):
    """ Process AS0 auto-mode setting """
    asm = _make_dat(stx[0:11], 4)   # 11-bit freq
    a1 = int(stx[11])               # 4-bit mode
    a2 = int(stx[12])               # 2-bit data
    asm += chr((a1 << 4) | (a2 << 2))
    return asm


def my_val_list(setting, opts, obj, atrb, fix=0, ndx=-1):
    """Callback:from ValueList. Set the integer index."""
    # This function is here to be available to get_mem and get_set
    # fix is optional additive offset to the list index
    # ndx is optional obj[ndx] array index
    value = opts.index(str(setting.value))
    value += fix
    if ndx >= 0:    # indexed obj
        setattr(obj[ndx], atrb, value)
    else:
        setattr(obj, atrb, value)
    return


def _read_settings(radio):
    """ Continue filling memory map"""
    global MEMSEL
    # setc: the list of CAT commands for downloaded settings
    # Block parameters first. In the exact order of MEM_FORMAT
    setc = radio.SETC
    setc.extend(radio.EX)  # Menu A EX params
    setc.extend(radio.EX)  # Menu B
    status = chirp_common.Status()
    status.cur = 0
    status.max = 32 + 50 + 8 + 11 + 39 + 39
    status.msg = "Reading Settings..."
    radio.status_fn(status)

    setts = ""
    nc = 0
    for cmc in setc:
        skipme = False
        argx = ""           # Extended arguments
        if cmc == "AS0":
            skipme = True   # flag to disable further processing
            for ix in range(32):        # 32 AS params
                result0 = command(radio.pipe, cmc, 19, W8S,
                                  "%02i" % ix)
                xc = len(cmc) + 2
                result0 = result0[xc:]
                setts += _sets_asf(result0)
                nc += 1
                status.cur = nc
                radio.status_fn(status)
        elif cmc == "SS":
            skipme = True
            for ix in range(10):     # 10 chans
                for nx in range(5):     # 5 spots
                    result0 = command(radio.pipe, cmc, 16, W8S,
                                      "%1i%1i" % (ix, nx))
                    setts += _make_dat(result0[4:], 4)
                    nc += 1
                    status.cur = nc
                    radio.status_fn(status)
        elif cmc == "EQ":
            skipme = True
            for ix in range(8):
                result0 = command(radio.pipe, cmc, 6, W8S, "0%1i"
                                  % ix)   # Tx eq
                setts += chr(int(result0[4:]))    # 'EQ13x", want the x
                result0 = command(radio.pipe, cmc, 6, W8S, "1%1i"
                                  % ix)   # Rx eq
                setts += chr(int(result0[4:]))
                nc += 1
                status.cur = nc
                radio.status_fn(status)
        elif ((not radio.SG) and (cmc == "EX087")) \
                or (radio.SG and (cmc == "EX001")):
            result0 = command(radio.pipe, cmc, 9, W8S, "0000")
            result0 += read_str(radio)    # Read pwron message
            result0 = result0[8:]
            nx = len(result0)
            for ix in range(8):
                if ix < nx:
                    sx = result0[ix]    # may need to test valid char
                    setts += sx
                else:
                    setts += chr(0)
            skipme = True
            nc += 1
            status.cur = nc
            radio.status_fn(status)
        elif (cmc == "MF0") or (cmc == "MF1"):
            result0 = command(radio.pipe, cmc, 0, W8S)
            skipme = True   # cmd only, no response
        else:   # issue the cmc cmd as-is with argx
            if str(cmc).startswith("EX"):
                argx = "0000"
            result0 = command(radio.pipe, cmc, 0, W8S, argx)
            result0 = read_str(radio)    # various length responses
            # strip the cmd echo
            xc = len(cmc)
            result0 = result0[xc:]
        # Cmd has been sent, process the result
        if cmc == "FV":      # all chars
            skipme = True
            setts += result0
        elif cmc == "AN":    # Antenna selection has 3 values
            skipme = True
            setts += _sets_val(result0, 3, 2)   # store as 2-bits each
        elif (cmc == "FA") or (cmc == "FB"):    # Response is 11-bit frq
            skipme = True
            setts += _make_dat(result0, 4)   # 11-bit freq
        elif (cmc == "MF0") or (cmc == "MF1"):  # No stored response
            skipme = True
        # Generic single byte processing
        if not skipme:
            setts += chr(int(result0))
        if cmc == "MF":     # Save the initial Menu selection
            MEMSEL = int(result0)
        nc += 1
        status.cur = nc
        radio.status_fn(status)
    setts += radio.MODEL.ljust(9)
    # Now set the inidial menu selection back
    result0 = command(radio.pipe, "MF", 0, W8L, "%1i" % MEMSEL)
    # And the original Beep Volume
    result0 = command(radio.pipe, "EX0050000%2i" % BEEPVOL, 0, W8L)
    return setts


def _write_mem(radio):
    """ Send MW commands for each channel """
    global BEEPVOL
    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = radio._upper + 20  # 10 P chans and 10 EXT
    status.msg = "Writing Channel Memory"
    radio.status_fn(status)

    result0 = command(radio.pipe, "EX0050000", 9, W8S)
    result0 += read_str(radio)
    BEEPVOL = int(result0[6:])
    result0 = command(radio.pipe, "EX005000000", 0, W8L)   # Silence beeps

    for chn in range(0, (radio._upper + 21)):   # Loop stops at +20
        _mem = radio._memobj.ch_mem[chn]
        cmx = "MW0 %02i" % chn
        if chn > 99:
            cmx = "MW0%03i" % chn
        stm = cmx + radio._make_base_spec(_mem, _mem.rxfreq)
        result0 = command(radio.pipe, stm, 0, W8L)     # No response
        cmx = "MW1 %02i" % chn
        if chn > 99:
            cmx = "MW1%03i" % chn
        stm = cmx + radio._make_base_spec(_mem, _mem.txfreq)
        if _mem.rxfreq > 0:         # Dont write MW1 if empty
            result0 = command(radio.pipe, stm, 0, W8L)
        # UI Update
        status.cur = chn
        radio.status_fn(status)
    return


def _write_sets(radio):
    """ Send settings and Menu a/b """
    status = chirp_common.Status()
    status.cur = 0
    status.max = 187   # Total to send
    status.msg = "Writing Settings"
    radio.status_fn(status)
    # Define mem struct shortcuts
    _sets = radio._memobj.settings
    _asf = radio._memobj.asf
    _ssf = radio._memobj.ssf
    _eqx = radio._memobj.eqx
    _mex = radio._memobj.exset
    snx = 0     # Settings status counter
    stlen = 0   # No response count
    # Send 32 AS
    for ix in range(32):
        scm = "AS0%02i%011i%1i%1i" % (ix, _asf[ix].asfreq,
                                      _asf[ix].asmode, _asf[ix].asdata)
        result0 = command(radio.pipe, scm, stlen, W8S)
        snx += 1
        status.cur = snx
        radio.status_fn(status)
    # Send 50 SS
    for ix in range(10):
        for kx in range(5):
            nx = ix * 5 + kx
            scm = "SS%1i%1i%011i" % (ix, kx, _ssf[nx].ssfreq)
            result0 = command(radio.pipe, scm, stlen, W8S)
            snx += 1
            status.cur = snx
            radio.status_fn(status)
    # Send 16 EQ
    for ix in range(8):
        scm = "EQ0%1i%1i" % (ix, _eqx[ix].txeq)
        result0 = command(radio.pipe, scm, stlen, W8S)
        scm = "EQ1%1i%1i" % (ix, _eqx[ix].rxeq)
        result0 = command(radio.pipe, scm, stlen, W8S)
        snx += 2
        status.cur = snx
        radio.status_fn(status)
    # Send 11 thingies
    scm = "AG0%03i" % _sets.ag
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "AN%1i%1i%1i" % (_sets.an1, _sets.an2, _sets.an3)
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "FA%011i" % _sets.fa
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "FB%011i" % _sets.fb
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "MG%03i" % _sets.mg
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "PC%03i" % _sets.pc
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "RG%03i" % _sets.rg
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "TP%03i" % _sets.tp
    result0 = command(radio.pipe, scm, stlen, W8S)
    scm = "MF0"   # Select menu A/B
    result0 = command(radio.pipe, scm, stlen, W8S)
    snx += 11
    status.cur = snx
    radio.status_fn(status)
    # Send 38 Menu A EX
    setc = radio.EX     # list of EX cmds
    for ix in range(2):
        for cmx in setc:
            if str(cmx)[0:2] == "MF":
                scm = cmx
            else:       # The EX cmds
                # Test for the power-on string
                if (radio.SG and cmx == "EX001") or \
                        ((not radio.SG) and cmx == "EX087"):
                    scm = cmx + "0000"
                    for chx in _mex[ix].ex001:  # Both get string here
                        scm += chr(chx)
                    scm = scm.strip()
                    scm = scm.strip(chr(0))     # in case any got thru
                # Now for the other EX cmds
                else:
                    if radio.SG:
                        scm = "%s0000%i" % (cmx, getattr(_mex[ix],
                                            cmx.lower()))
                    else:   # Gotta use the cross reference dict for cmd
                        scm = "%s0000%i" % (cmx, getattr(_mex[ix],
                                            radio.EX_X[cmx].lower()))
            result0 = command(radio.pipe, scm, stlen, W8S)
            snx += 1
            status.cur = snx
            radio.status_fn(status)
    # Now set the inidial menu selection back
    result0 = command(radio.pipe, "MF", 0, W8L, "%1i" % _sets.mf)
    # And the original Beep Volume
    result0 = command(radio.pipe, "EX0050000%2i" % BEEPVOL, 0, W8L)
    return


@directory.register
class TS590Radio(chirp_common.CloneModeRadio):
    """Kenwood TS-590"""
    VENDOR = "Kenwood"
    MODEL = "TS-590SG_CloneMode"
    ID = "ID023;"
    SG = True
    # Settings read/write cmd sequence list
    SETC = ["AS0", "SS", "EQ", "AG0", "AN", "FA", "FB",
            "FV", "MF", "MG", "PC", "RG", "TP", "MF0"]
    # This is the TS-590SG MENU A/B read_settings parameter tuple list
    # The order is mandatory; to match the Mem_Format sequence
    EX = ["EX001", "EX002", "EX003", "EX005", "EX006", "EX007",
          "EX008", "EX009", "EX010", "EX011", "EX012", "EX013", "EX016",
          "EX017", "EX018", "EX019", "EX021", "EX022", "EX023", "EX024",
          "EX025", "EX026", "EX054", "EX055", "EX076", "EX077", "EX087",
          "EX088", "EX089", "EX090", "EX091", "EX092", "EX093", "EX094",
          "EX095", "EX096", "EX097", "EX098", "EX099", "MF1"]
    # EX menu settings label dictionary. Key is the EX number
    EX_LBL = {2: " Display brightness",
              1: " Power-On message",
              3: " Backlight color",
              5: " Beep volume",
              6: " Sidetone volume",
              7: " Message playback volume",
              8: " Voice guide volume",
              9: " Voice guide speed",
              10: " Voice guide language",
              11: " Auto Announcement",
              12: " MHz step",
              13: " Tuning control adj rate (Hz)",
              16: " SSB tune step (kHz)",
              17: " CW/FSK tune step (kHz)",
              18: " AM tune step (kHz)",
              19: " FM tune step (kHz)",
              21: " Max number of Quick Mem chans",
              22: " Temporary MR Chan freq allowed",
              23: " Program Scan slowdown",
              24: " Program Scan slowdown range (Hz)",
              25: " Program Scan hold",
              26: " Scan Resume method",
              54: " TX Power fine adjust",
              55: " Timeout timer (Secs)",
              76: " Data VOX",
              77: " Data VOX delay (x30 mSecs)",
              87: " Panel PF-A function",
              88: " Panel PF-B function",
              89: " RIT key function",
              90: " XIT key function",
              91: " CL key function",
              92: " Front panel MULTI/CH key (non-CW mode)",
              93: " Front panel MULTI/CH key (CW mode)",
              94: " MIC PF1 function",
              95: " MIC PF2 function",
              96: " MIC PF3 function",
              97: " MIC PF4 function",
              98: " MIC PF (DWN) function",
              99: " MIC PF (UP) function"}

    BAUD_RATE = 115200
    _upper = 99

    # Special Channels Declaration
    # WARNING Indecis are hard wired in get/set_memory code !!!
    # Channels print in + increasing index order
    SPECIAL_MEMORIES = {"EXT 0": 110,
                        "EXT 1": 111,
                        "EXT 2": 112,
                        "EXT 3": 113,
                        "EXT 4": 114,
                        "EXT 5": 115,
                        "EXT 6": 116,
                        "EXT 7": 117,
                        "EXT 8": 118,
                        "EXT 9": 119}

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.can_odd_split = False
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_dtcs = False
        rf.has_dtcs_polarity = False
        rf.has_name = True
        rf.has_settings = True
        rf.has_offset = True
        rf.has_mode = True
        rf.has_tuning_step = False  # Not in mem chan
        rf.has_nostep_tuning = True     # Radio accepts any entered freq
        rf.has_cross = True
        rf.has_comment = False
        rf.memory_bounds = (0, self._upper)
        rf.valid_bands = [(30000, 24999999), (25000000, 59999999)]
        rf.valid_characters = chirp_common.CHARSET_UPPER_NUMERIC + "*+-/"
        rf.valid_duplexes = TS590_DUPLEX
        rf.valid_modes = TS590_MODES
        rf.valid_skips = TS590_SKIP
        rf.valid_tuning_steps = [0.5, 1.0, 2.5, 5.0, 6.25, 10.0, 12.5,
                                 15.0, 20.0, 25.0, 30.0, 50.0, 100.0]
        rf.valid_tmodes = ["", "Tone", "TSQL", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone"]
        rf.valid_name_length = 8    # 8 character channel names
        rf.valid_special_chans = sorted(self.SPECIAL_MEMORIES.keys())

        return rf

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.info = _(
            "Click on the \"Special Channels\" toggle-button of the memory\n"
            "editor to see/set the EXT channels. P-VFO channels 100-109\n"
            "are considered Settings.\n"
            "Only a subset of the over 200 available radio settings\n"
            "are supported in this release.\n"
            "Ignore the beeps from the radio on upload and download.\n")
        rp.pre_download = _(
            "Follow these instructions to download the radio memory:\n"
            "1 - Connect your interface cable\n"
            "2 - Radio > Download from radio: DO NOT mess with the radio\n"
            "during download!\n"
            "3 - Disconnect your interface cable\n")
        rp.pre_upload = _(
            "Follow these instructions to upload the radio memory:\n"
            "1 - Connect your interface cable\n"
            "2 - Radio > Upload to radio: DO NOT mess with the radio\n"
            "during upload!\n"
            "3 - Disconnect your interface cable\n")
        return rp

    def sync_in(self):
        """Download from radio"""
        try:
            _connect_radio(self)
            data = _read_mem(self)
            data += _read_settings(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

        self._mmap = memmap.MemoryMap(data)
        self.process_mmap()
        return

    def sync_out(self):
        """Upload to radio"""
        try:
            _connect_radio(self)
            _write_mem(self)
            _write_sets(self)
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        return

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)
        return

    def get_memory(self, number):
        """Convert raw channel data (_mem) into UI columns (mem)"""
        mem = chirp_common.Memory()
        mem.extra = RadioSettingGroup("extra", "Extra")
        if isinstance(number, str):
            mem.name = number   # Spcl chns 1st var
            mem.number = self.SPECIAL_MEMORIES[number]
            _mem = self._memobj.ch_mem[mem.number]
        else:       # Normal mem chans and VFO edges
            if number > 99 and number < 110:
                return          # Don't show VFO edges as mem chans
            _mem = self._memobj.ch_mem[number]
            mem.number = number
            mnx = ""
            for char in _mem.name:
                mnx += chr(char)
            mem.name = mnx.strip()
            mem.name = mem.name.upper()
        # From here on is common to both types
        if _mem.rxfreq == 0:
            _mem.txfreq = 0
            mem.empty = True
            mem.freq = 0
            mem.mode = "LSB"
            mem.offset = 0
            return mem
        mem.empty = False
        mem.freq = int(_mem.rxfreq)
        mem.duplex = TS590_DUPLEX[0]    # None by default
        mem.offset = 0
        if _mem.rxfreq < _mem.txfreq:   # + shift
            mem.duplex = TS590_DUPLEX[2]
            mem.offset = _mem.txfreq - _mem.rxfreq
        if _mem.rxfreq > _mem.txfreq:   # - shift
            mem.duplex = TS590_DUPLEX[1]
            mem.offset = _mem.rxfreq - _mem.txfreq
        if _mem.txfreq == 0:
            # leave offset alone, or run_tests will bomb
            mem.duplex = TS590_DUPLEX[0]
        mx = _mem.xmode - 1     # CAT modes start at 1
        if _mem.xmode == 9:     # except CAT FSK-R is 9, there is no 8
            mx = 7
        if _mem.data:       # LSB+Data= 8, USB+Data= 9, FM+Data= 10
            if _mem.xmode == 1:     # CAT LSB
                mx = 8
            elif _mem.xmode == 2:   # CAT USB
                mx = 9
            elif _mem.xmode == 4:   # CAT FM
                mx = 10
        mem.mode = TS590_MODES[mx]
        mem.tmode = ""
        mem.cross_mode = "Tone->Tone"
        mem.ctone = TS590_TONES[_mem.ctone]
        mem.rtone = TS590_TONES[_mem.rtone]
        if _mem.tmode == 1:
            mem.tmode = "Tone"
        elif _mem.tmode == 2:
            mem.tmode = "TSQL"
        elif _mem.tmode == 3:
            mem.tmode = "Cross"
        mem.skip = TS590_SKIP[_mem.skip]

        # Channel Extra settings: Only Boolean & List methods, no call-backs
        options = ["Wide", "Narrow"]
        rx = RadioSettingValueList(options, options[_mem.fmnrw])
        # NOTE: first param of RadioSetting is the object attribute name
        rset = RadioSetting("fmnrw", "FM mode", rx)
        rset.set_apply_callback(my_val_list, options, _mem, "fmnrw")
        mem.extra.append(rset)

        options = ["Filter A", "Filter B"]
        rx = RadioSettingValueList(options, options[_mem.filter])
        rset = RadioSetting("filter", "Filter A/B", rx)
        rset.set_apply_callback(my_val_list, options, _mem, "filter")
        mem.extra.append(rset)

        return mem

    def set_memory(self, mem):
        """Convert UI column data (mem) into MEM_FORMAT memory (_mem)"""
        _mem = self._memobj.ch_mem[mem.number]
        if mem.empty:
            _mem.rxfreq = 0
            _mem.txfreq = 0
            _mem.xmode = 0
            _mem.data = 0
            _mem.tmode = 0
            _mem.rtone = 0
            _mem.ctone = 0
            _mem.filter = 0
            _mem.skip = 0
            _mem.fmnrw = 0
            _mem.name = "        "
            return

        if mem.number > self._upper:    # Specials: No Name changes
            ix = 0
            # LOG.warning("Special Chan set_mem @ %i" % mem.number)
        else:
            nx = len(mem.name)
            for ix in range(8):
                if ix < nx:
                    _mem.name[ix] = mem.name[ix].upper()
                else:
                    _mem.name[ix] = " "    # assignment needs 8 chrs
        _mem.rxfreq = mem.freq
        _mem.txfreq = 0
        if mem.duplex == "+":
            _mem.txfreq = mem.freq + mem.offset
        if mem.duplex == "-":
            _mem.txfreq = mem.freq - mem.offset
        ix = TS590_MODES.index(mem.mode)
        _mem.data = 0
        _mem.xmode = ix + 1     # stored as CAT values, LSB= 1
        if ix == 7:     # FSK-R
            _mem.xmode = 9      # There is no CAT 8
        if ix > 7:      # a Data mode
            _mem.data = 1
            if ix == 8:
                _mem.xmode = 1      # LSB
            elif ix == 9:
                _mem.xmode = 2      # USB
            elif ix == 10:
                _mem.xmode = 4      # FM
        _mem.tmode = 0
        _mem.rtone = TS590_TONES.index(mem.rtone)
        _mem.ctone = TS590_TONES.index(mem.ctone)
        if mem.tmode == "Tone":
            _mem.tmode = 1
        if mem.tmode == "TSQL":
            _mem.tmode = 2
        if mem.tmode == "Cross" or mem.tmode == "Tone->Tone":
            _mem.tmode = 3
        _mem.skip = 0
        if mem.skip == "S":
            _mem.skip = 1
        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)
        return

    def _parse_mem_spec(self, spec0, spec1):
        """ Extract ascii memory parameters; build data string """
        # spec0 is simplex result, spec1 is split
        # pad string so indexes match Kenwood docs
        spec0 = "x" + spec0  # match CAT document 1-based description
        ix = len(spec0)
        # _pxx variables are STRINGS
        _p1 = spec0[3]       # P1    Split Specification
        _p3 = spec0[4:7]     # P3    Memory Channel
        _p4 = spec0[7:18]    # P4    Frequency
        _p5 = spec0[18]      # P5    Mode
        _p6 = spec0[19]      # P6    Data Mode
        _p7 = spec0[20]      # P7    Tone Mode
        _p8 = spec0[21:23]   # P8    Tone Frequency Index
        if _p8 == "00":      # Can't be 0 at upload
            _p8 = "08"
        _p9 = spec0[23:25]   # P9    CTCSS Frequency Index
        if _p9 == "00":
            _p9 = "08"
        _p10 = spec0[25:28]  # P10   Always 0
        _p11 = spec0[28]     # P11   Filter A/B
        _p12 = spec0[29]     # P12   Always 0
        _p13 = spec0[30:38]  # P13   Always 0
        _p14 = spec0[38:40]  # P14   FM Mode
        _p15 = spec0[40]     # P15   Chan Lockout (Skip)
        _p16 = spec0[41:49]  # P16   Max 8-Char Name if assigned

        spec1 = "x" + spec1
        _p4s = int(spec1[7:18])  # P4: Offset freq

        datm = ""   # Fill in MEM_FORMAT sequence
        datm += _make_dat(_p4, 4)   # rxreq: u32, 4 bytes/chars
        datm += _make_dat(_p4s, 4)  # tx freq
        v1 = int(_p5) << 4          # xMode: 0-9, upper 4 bits
        v1 = v1 | (int(_p6) << 2)   # Data: 0/1
        v1 = v1 | int(_p7)          # Tmode: 0-3
        datm += chr(v1)
        datm += chr(int(_p8))       # rtone: 00-42
        datm += chr(int(_p9))       # ctone
        v1 = int(_p11) << 7         # Filter A/B 1 bit msb
        v1 = v1 | (int(_p14) << 6)  # fmwide: 1 bit
        v1 = v1 | (int(_p15) << 5)  # skip: 1 bit
        datm += chr(v1)
        v1 = len(_p16)
        for ix in range(8):
            if ix < v1:
                datm += _p16[ix]
            else:
                datm += " "

        return datm

    def _make_base_spec(self, mem, freq):
        """ Generate memory channel parameter string """
        spec = "%011i%1i%1i%1i%02i%02i000%1i0000000000%02i%1i%s" \
            % (freq, mem.xmode, mem.data, mem.tmode, mem.rtone,
                mem.ctone, mem.filter, mem.fmnrw, mem.skip, mem.name)

        return spec.strip()

    def get_settings(self):
        """Translate the MEM_FORMAT structs into settings in the UI"""
        # Define mem struct write-back shortcuts
        _sets = self._memobj.settings
        _asf = self._memobj.asf
        _ssf = self._memobj.ssf
        _eqx = self._memobj.eqx
        _mex = self._memobj.exset
        _chm = self._memobj.ch_mem
        basic = RadioSettingGroup("basic", "Basic Settings")
        pvfo = RadioSettingGroup("pvfo", "VFO Band Edges")
        mena = RadioSettingGroup("mena", "Menu A")
        menb = RadioSettingGroup("menb", "Menu B")
        equ = RadioSettingGroup("equ", "Equalizers")
        amode = RadioSettingGroup("amode", "Auto Mode")
        ssc = RadioSettingGroup("ssc", "Slow Scan")
        group = RadioSettings(basic, pvfo, mena, menb, equ, amode, ssc)

        mhz1 = 1000000.
        nsg = not self.SG
        if nsg:     # Make reverse EX_X dictionary
            x_ex = {v: k for k, v in self.EX_X.items()}

        # Callback functions
        def _my_readonly(setting, obj, atrb):
            """NOP callback, prevents writing the setting"""
            vx = 0
            return

        def my_adjraw(setting, obj, atrb, fix=0, ndx=-1):
            """Callback for Integer add or subtract fix from value."""
            vx = int(str(setting.value))
            value = vx + int(fix)
            if value < 0:
                value = 0
            if ndx < 0:
                setattr(obj, atrb, value)
            else:
                setattr(obj[ndx], atrb, value)
            return

        def my_mhz_val(setting, obj, atrb, ndx=-1):
            """ Callback to set freq back to Hz """
            vx = float(str(setting.value))
            vx = int(vx * mhz1)
            if ndx < 0:
                setattr(obj, atrb, vx)
            else:
                setattr(obj[ndx], atrb, vx)
            return

        def my_bool(setting, obj, atrb, ndx=-1):
            """ Callback to properly set boolean """
            # set_settings is not setting [indexed] booleans???
            vx = 0
            if str(setting.value) == "True":
                vx = 1
            if ndx < 0:
                setattr(obj, atrb, vx)
            else:
                setattr(obj[ndx], atrb, vx)
            return

        def my_asf_mode(setting, obj, nx=0):
            """ Callback to extract mode and create asmode, asdata """
            v1 = TS590_MODES.index(str(setting.value))
            v2 = 0      # asdata
            vx = v1 + 1     # stored as CAT values, same as xmode
            if v1 == 7:
                vx = 9
            if v1 > 7:      # a Data mode
                v2 = 1
                if v1 == 8:
                    vx = 1      # LSB
                elif v1 == 9:
                    vx = 2      # USB
                elif v1 == 10:
                    vx = 4      # FM
            setattr(obj[nx], "asdata", v2)
            setattr(obj[nx], "asmode", vx)
            return

        def my_fnctns(setting, obj, ndx, atrb):
            """ Filter only valid key function assignments """
            vx = int(str(setting.value))
            if self.SG:
                vmx = 210
                if (vx > 99 and vx < 120) or (vx > 170 and vx < 200):
                    raise errors.RadioError(" %i Change Ignored for %s."
                                            % (vx, atrb))
                    return  # not valid, ignored
            else:
                vmx = 208
                if (vx > 87 and vx < 100) or (vx > 134 and vx < 200):
                    raise errors.RadioError(" %i Change Ignored for %s."
                                            % (vx, atrb))
                    return
            if vx > vmx:
                vx = 255       # Off
            setattr(obj[ndx], atrb, vx)
            return

        def my_labels(kx):      # nsg and x_ex defined above
            lbl = "%03i:" % kx      # SG EX number
            if nsg:
                lbl = x_ex["EX%03i" % kx][2:] + ":"    # S-model EX num
            lbl += self.EX_LBL[kx]      # and the label to match
            return lbl

        # ===== BASIC GROUP =====
        sx = ""
        for i in range(4):
            sx += chr(_sets.fv[i])
        rx = RadioSettingValueString(0, 4, sx)
        rset = RadioSetting("settings.fv", "FirmwareVersion", rx)
        rset.set_apply_callback(_my_readonly, _sets, "fv")
        basic.append(rset)

        rx = RadioSettingValueInteger(0, 255, _sets.ag + 1)
        rset = RadioSetting("settings.ag", "AF Gain", rx)
        rset.set_apply_callback(my_adjraw, _sets, "ag", -1)
        basic.append(rset)

        rx = RadioSettingValueInteger(0, 255, _sets.rg + 1)
        rset = RadioSetting("settings.rg", "RF Gain", rx)
        rset.set_apply_callback(my_adjraw, _sets, "rg", -1)
        basic.append(rset)

        options = ["ANT1", "ANT2"]
        # CAUTION: an1 has value of 1 or 2
        rx = RadioSettingValueList(options, options[_sets.an1 - 1])
        rset = RadioSetting("settings.an1", "Antenna Selected", rx)
        # Add 1 to the changed value. S/b 1/2
        rset.set_apply_callback(my_val_list, options, _sets, "an1", 1)
        basic.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.an2))
        rset = RadioSetting("settings.an2", "Recv Antenna is used", rx)
        basic.append(rset)

        rx = RadioSettingValueBoolean(bool(_sets.an3))
        rset = RadioSetting("settings.an3", "Drive Out On", rx)
        basic.append(rset)

        rx = RadioSettingValueInteger(0, 100, _sets.mg)
        rset = RadioSetting("settings.mg", "Microphone gain", rx)
        basic.append(rset)

        nx = 5      # Coarse step
        if bool(_mex[0].ex054):   # Power Fine enabled in menu A
            nx = 1
        vx = _sets.pc       # Trap invalid values from run_tests.py
        if vx < 5:
            vx = 5
        rx = RadioSettingValueInteger(5, 100, vx, nx)
        sx = "TX Output power (Watts)"
        rset = RadioSetting("settings.pc", sx, rx)
        basic.append(rset)

        vx = _sets.tp
        rx = RadioSettingValueInteger(5, 100, vx, nx)
        sx = "TX Tuning power (Watts)"
        rset = RadioSetting("settings.tp", sx, rx)
        basic.append(rset)

        val = _sets.fa / mhz1       # Allow Rx freq range
        rx = RadioSettingValueFloat(0.3, 60.0, val, 0.001, 3)
        sx = "VFO-A Frequency (MHz)"
        rset = RadioSetting("settings.fa", sx, rx)
        rset.set_apply_callback(my_mhz_val, _sets, "fa")
        basic.append(rset)

        val = _sets.fb / mhz1
        rx = RadioSettingValueFloat(0.3, 60.0, val, 0.001, 3)
        sx = "VFO-B Frequency (MHz)"
        rset = RadioSetting("settings.fb", sx, rx)
        rset.set_apply_callback(my_mhz_val, _sets, "fb")
        basic.append(rset)

        options = ["Menu A", "Menu B"]
        rx = RadioSettingValueList(options, options[_sets.mf])
        sx = "Menu Selected"
        rset = RadioSetting("settings.mf", sx, rx)
        rset.set_apply_callback(my_val_list, options, _sets, "mf")
        basic.append(rset)

        # ==== VFO Edges Group ================

        for mx in range(100, 110):
            val = _chm[mx].rxfreq / mhz1
            if val < 1.8:       # Many operators never use this
                val = 1.8       # So default is 0.0
            rx = RadioSettingValueFloat(1.8, 54.0, val, 0.001, 3)
            sx = "VFO-Band %i lower limit (MHz)" % (mx - 100)
            rset = RadioSetting("ch_mem.rxfreq/%d" % mx, sx, rx)
            rset.set_apply_callback(my_mhz_val, _chm, "rxfreq", mx)
            pvfo.append(rset)

            val = _chm[mx].txfreq / mhz1
            if val < 1.8:
                val = 54.0
            rx = RadioSettingValueFloat(1.8, 54.0, val, 0.001, 3)
            sx = "    VFO-Band %i upper limit (MHz)" % (mx - 100)
            rset = RadioSetting("ch_mem.txfreq/%d" % mx, sx, rx)
            rset.set_apply_callback(my_mhz_val, _chm, "txfreq", mx)
            pvfo.append(rset)

            kx = _chm[mx].xmode
            options = ["None", "LSB", "USB", "CW", "FM", "AM", "FSK",
                       "CW-R", "N/A", "FSK-R"]
            rx = RadioSettingValueList(options, options[kx])
            sx = "    VFO-Band %i Tx/Rx Mode" % (mx - 100)
            rset = RadioSetting("ch_mem.xmode/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _chm,
                                    "xmode", 0, mx)
            pvfo.append(rset)

        # ==== Menu A/B Group =================

        for mx in range(2):      # A/B index
            sx = ""
            for i in range(8):
                if int(_mex[mx].ex001[i]) != 0:
                    sx += chr(_mex[mx].ex001[i])
            sx = sx.strip()
            rx = RadioSettingValueString(0, 8, sx)
            sx = my_labels(1)     # Proper label for EX001
            rset = RadioSetting("exset.ex001/%d" % mx, sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            sx = my_labels(2)
            rx = RadioSettingValueInteger(0, 6, _mex[mx].ex002)
            rset = RadioSetting("exset.ex002", sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            nx = 2
            if self.SG:
                nx = 10
            vx = _mex[mx].ex003 + 1     # radio rtns 0-9
            rx = RadioSettingValueInteger(1, nx, vx)
            sx = my_labels(3)
            rset = RadioSetting("exset.ex003", sx, rx)
            rset.set_apply_callback(my_adjraw, _mex, "ex003", -1, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            nx = 9
            if self.SG:
                nx = 20
            rx = RadioSettingValueInteger(0, nx, _mex[mx].ex005)
            sx = my_labels(5)
            rset = RadioSetting("exset.ex005", sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            sx = my_labels(6)
            rx = RadioSettingValueInteger(0, nx, _mex[mx].ex006)
            rset = RadioSetting("exset.ex006", sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            sx = my_labels(7)
            rx = RadioSettingValueInteger(0, nx, _mex[mx].ex007)
            rset = RadioSetting("exset.ex007", sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            nx = 7
            if self.SG:
                nx = 20
            sx = my_labels(8)
            rx = RadioSettingValueInteger(0, nx, _mex[mx].ex008)
            rset = RadioSetting("exset.ex008", sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            sx = my_labels(9)
            rx = RadioSettingValueInteger(0, 4, _mex[mx].ex009)
            rset = RadioSetting("exset.ex009", sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["English", "Japanese"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex010])
            sx = my_labels(10)
            rset = RadioSetting("exset.ex010/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex010", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["Off", "1", "2"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex011])
            sx = my_labels(11)
            rset = RadioSetting("exset.ex011/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex011", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["0.1", "0.5", "1.0"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex012])
            sx = my_labels(12)
            rset = RadioSetting("exset.ex012", sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex012", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["250", "500", "1000"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex013])
            sx = my_labels(13)
            rset = RadioSetting("exset.ex013/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex013", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            # S and SG have different ranges for control steps
            options = ["0.5", "1.0", "2.5", "5.0", "10.0"]
            if self.SG:
                options = ["Off", "0.5", "0.5", "1.0", "2.5",
                           "5.0", "10.0"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex016])
            sx = my_labels(16)
            if nsg:
                sx = "014: Tuning step for SSB/CW/FSK (kHz)"
            rset = RadioSetting("exset.ex016/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex016", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            if self.SG:       # this setting only for SG
                rx = RadioSettingValueList(options,
                                           options[_mex[mx].ex017])

                sx = my_labels(17)
                rset = RadioSetting("exset.ex017/%d" % mx, sx, rx)
                rset.set_apply_callback(my_val_list, options, _mex,
                                        "ex017", 0, mx)
                if mx == 0:
                    mena.append(rset)
                else:
                    menb.append(rset)

            options = ["Off", "5.0", "6.25", "10.0", "12.5", "15.0",
                       "20.0", "25.0", "30.0", "50.0", "100.0"]
            if self.SG:
                options.remove("Off")
            rx = RadioSettingValueList(options, options[_mex[mx].ex018])
            sx = my_labels(18)
            rset = RadioSetting("exset.ex018/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex018", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueList(options, options[_mex[mx].ex019])
            sx = my_labels(19)
            rset = RadioSetting("exset.ex019/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex019", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["3", "5", "10"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex021])
            sx = my_labels(21)
            rset = RadioSetting("exset.ex021/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex021", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueBoolean(bool(_mex[mx].ex022))
            sx = my_labels(22)
            rset = RadioSetting("exset.ex022/%d" % mx, sx, rx)
            rset.set_apply_callback(my_bool, _mex, "ex022", mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueBoolean(bool(_mex[mx].ex023))
            sx = my_labels(23)
            rset = RadioSetting("exset.ex023/%d" % mx, sx, rx)
            rset.set_apply_callback(my_bool, _mex, "ex023", mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["100", "200", "300", "400", "500"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex024])
            sx = my_labels(24)
            rset = RadioSetting("exset.ex024/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex024", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueBoolean(bool(_mex[mx].ex025))
            sx = my_labels(25)
            rset = RadioSetting("exset.ex025/%d" % mx, sx, rx)
            rset.set_apply_callback(my_bool, _mex, "ex025", mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["TO", "CO"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex026])
            sx = my_labels(26)
            rset = RadioSetting("exset.ex026/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex026", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueBoolean(bool(_mex[mx].ex054))
            sx = my_labels(54)
            rset = RadioSetting("exset.ex054/%d" % mx, sx, rx)
            rset.set_apply_callback(my_bool, _mex, "ex054", mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            options = ["Off", "3", "5", "10", "20", "30"]
            rx = RadioSettingValueList(options, options[_mex[mx].ex055])
            sx = my_labels(55)
            rset = RadioSetting("exset.ex055/%d" % mx, sx, rx)
            rset.set_apply_callback(my_val_list, options, _mex,
                                    "ex055", 0, mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueBoolean(bool(_mex[mx].ex076))
            sx = my_labels(76)
            rset = RadioSetting("exset.ex076/%d" % mx, sx, rx)
            rset.set_apply_callback(my_bool, _mex, "ex076", mx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 100, _mex[mx].ex077, 5)
            sx = my_labels(77)
            rset = RadioSetting("exset.ex077/%d" % mx, sx, rx)
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex087)
            sx = my_labels(87)
            rset = RadioSetting("exset.ex087/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex087")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex088)
            sx = my_labels(88)
            rset = RadioSetting("exset.ex088/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex088")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            if self.SG:       # Next 5 settings not supported in 590S
                rx = RadioSettingValueInteger(0, 256, _mex[mx].ex089)
                sx = my_labels(89)
                rset = RadioSetting("exset.ex089/%d" % mx, sx, rx)
                rset.set_apply_callback(my_fnctns, _mex, mx, "ex089")
                if mx == 0:
                    mena.append(rset)
                else:
                    menb.append(rset)

                rx = RadioSettingValueInteger(0, 256, _mex[mx].ex090)
                sx = my_labels(90)
                rset = RadioSetting("exset.ex090/%d" % mx, sx, rx)
                rset.set_apply_callback(my_fnctns, _mex, mx, "ex090")
                if mx == 0:
                    mena.append(rset)
                else:
                    menb.append(rset)

                rx = RadioSettingValueInteger(0, 256, _mex[mx].ex091)
                sx = my_labels(91)
                rset = RadioSetting("exset.ex091/%d" % mx, sx, rx)
                rset.set_apply_callback(my_fnctns, _mex, mx, "ex091")
                if mx == 0:
                    mena.append(rset)
                else:
                    menb.append(rset)

                rx = RadioSettingValueInteger(0, 256, _mex[mx].ex092)
                sx = my_labels(92)
                rset = RadioSetting("exset.ex092/%d" % mx, sx, rx)
                rset.set_apply_callback(my_fnctns, _mex, mx, "ex092")
                if mx == 0:
                    mena.append(rset)
                else:
                    menb.append(rset)

                rx = RadioSettingValueInteger(0, 256, _mex[mx].ex093)
                sx = my_labels(93)
                rset = RadioSetting("exset.ex093/%d" % mx, sx, rx)
                rset.set_apply_callback(my_fnctns, _mex, mx, "ex093")
                if mx == 0:
                    mena.append(rset)
                else:
                    menb.append(rset)

            # Now both S and SG models
            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex094)
            sx = my_labels(94)
            rset = RadioSetting("exset.ex094/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex094")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex095)
            sx = my_labels(95)
            rset = RadioSetting("exset.ex095/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex095")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex096)
            sx = my_labels(96)
            rset = RadioSetting("exset.ex096/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex096")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex097)
            sx = my_labels(97)
            rset = RadioSetting("exset.ex097/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex097")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex098)
            sx = my_labels(98)
            rset = RadioSetting("exset.ex098/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex098")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)

            rx = RadioSettingValueInteger(0, 256, _mex[mx].ex099)
            sx = my_labels(99)
            rset = RadioSetting("exset.ex099/%d" % mx, sx, rx)
            rset.set_apply_callback(my_fnctns, _mex, mx, "ex099")
            if mx == 0:
                mena.append(rset)
            else:
                menb.append(rset)
        # End of for mx loop

        # ==== Auto Scan Params (amode) ==============
        for ix in range(32):
            val = _asf[ix].asfreq / mhz1
            rx = RadioSettingValueFloat(0.03, 60.0, val, 0.001, 3)
            rset = RadioSetting("asf.asfreq/%d" % ix,
                                "Scan %02i Freq (MHz)" % ix, rx)
            rset.set_apply_callback(my_mhz_val, _asf, "asfreq", ix)
            amode.append(rset)

            mx = _asf[ix].asmode - 1     # Same logic as xmode
            if _asf[ix].asmode == 9:
                mx = 7
            if _asf[ix].asdata:
                if _asf[ix].asmode == 1:
                    mx = 8
                elif _asf[ix].asmode == 2:
                    mx = 9
                elif _asf[ix].asmode == 4:
                    mx = 10
            rx = RadioSettingValueList(TS590_MODES, TS590_MODES[mx])
            rset = RadioSetting("asf.asmode/%d" % ix, "   Mode", rx)
            rset.set_apply_callback(my_asf_mode, _asf, ix)
            amode.append(rset)

        # ==== Slow Scan Settings ===
        for ix in range(10):        # Chans
            for nx in range(5):     # spots
                px = ((ix * 5) + nx)
                val = _ssf[px].ssfreq / mhz1
                stx = "      -   -   -    Slot %02i Freq (MHz)" % nx
                if nx == 0:
                    stx = "Slow Scan %02i, Slot 0 Freq (MHz" % ix
                rx = RadioSettingValueFloat(0, 54.0, val, 0.001, 3)
                rset = RadioSetting("ssf.ssfreq/%d" % px, stx, rx)
                rset.set_apply_callback(my_mhz_val, _ssf, "ssfreq", px)
                ssc.append(rset)

        # ==== Equalizer subgroup =====
        mohd = ["SSB", "SSB-DATA", "CW/CW-R", "FM", "FM-DATA", "AM",
                "AM-DATA", "FSK/FSK-R"]
        tcurves = ["Off", "HB1", "HB2", "FP", "BB1", "BB2",
                   "C", "U"]
        rcurves = ["Off", "HB1", "HB2", "FP", "BB1", "BB2",
                   "FLAT", "U"]
        for ix in range(8):
            rx = RadioSettingValueList(tcurves, tcurves[_eqx[ix].txeq])
            rset = RadioSetting("eqx.txeq/%d" % ix, "TX %s Equalizer"
                                % mohd[ix], rx)
            rset.set_apply_callback(my_val_list, tcurves, _eqx,
                                    "txeq", 0, ix)
            equ.append(rset)

            rx = RadioSettingValueList(rcurves, rcurves[_eqx[ix].rxeq])
            rset = RadioSetting("eqx.rxeq/%d" % ix, "RX %s Equalizer"
                                % mohd[ix], rx)
            rset.set_apply_callback(my_val_list, rcurves, _eqx,
                                    "rxeq", 0, ix)
            equ.append(rset)

        return group       # END get_settings()

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, fdata, fyle):
        """ Included to prevent 'File > New' error """
        return False


@directory.register
class TS590SRadio(TS590Radio):
    """ Kenwood TS-590S variant of the TS590 """
    VENDOR = "Kenwood"
    MODEL = "TS-590S_CloneMode"
    ID = "ID021;"
    SG = False
    # This is the equivalent Menu A/B list for the TS-590S
    # The equivalent S param is stored in the SG Mem_Format slot
    EX = ["EX087", "EX000", "EX001", "EX003", "EX004", "EX005",
          "EX006", "EX007", "EX008", "EX009", "EX010", "EX011", "EX014",
          "EX014", "EX015", "EX016", "EX017", "EX018", "EX019", "EX020",
          "EX021", "EX022", "EX048", "EX049", "EX069", "EX070", "EX079",
          "EX080", "EX080", "EX080", "EX080", "EX080", "EX080", "EX081",
          "EX082", "EX083", "EX084", "EX085", "EX086", "MF1"]
    # EX cross reference dictionary- key is S param, value is the SG
    EX_X = {"EX087": "EX001", "EX000": "EX002", "EX001": "EX003",
            "EX003": "EX005", "EX004": "EX006", "EX005": "EX007",
            "EX006": "EX008", "EX007": "EX009", "EX008": "EX010",
            "EX009": "EX011", "EX010": "EX012", "EX011": "EX013",
            "EX014": "EX016", "EX015": "EX018", "EX081": "EX094",
            "EX016": "EX019", "EX017": "EX021", "EX018": "EX022",
            "EX019": "EX023",
            "EX020": "EX024", "EX021": "EX025", "EX022": "EX026",
            "EX048": "EX054", "EX049": "EX055", "EX069": "EX076",
            "EX070": "EX077", "EX079": "EX087", "EX080": "EX088",
            "EX082": "EX095", "EX083": "EX096", "EX084": "EX097",
            "EX085": "EX098", "EX086": "EX099"}



# Copyright 2017 Pavel Milanes, CO7WT, <pavelmc@gmail.com>
#
# This driver is a community effort as I don't have the radio on my hands, so
# I was only the director of the orchestra, without the players this may never
# came true, so special thanks to the following hams for their contribution:
# - Henk van der Laan, PA3CQN
#       - Setting Discovery.
#       - Special channels for RELAY and EMERGENCY.
# - Harold Hankins
#       - Memory limits, testing & bug hunting.
# - Dmitry Milkov
#       - Testing & bug hunting.
# - Many others participants in the issue page on Chirp's site.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSetting, RadioSettingGroup, \
                RadioSettingValueBoolean, RadioSettingValueList, \
                RadioSettingValueInteger, RadioSettingValueFloat, \
                RadioSettings

import struct
import logging

LOG = logging.getLogger(__name__)

# A note about the memory in these radios
#
# The '9100' OEM software only manipulates the lower 0x0180 bytes on read/write
# operations as we know, the file generated by the OEM software IS NOT an exact
# eeprom image, it's a crude text file with a pseudo csv format
#
# Later investigations by Harold Hankins found that the eeprom extend up to 2k
# consistent with a hardware chip K24C16 a 2k x 8 bit serial eeprom

MEM_SIZE = 0x0800  # 2048 bytes
WRITE_SIZE = 0x0180  # 384 bytes
BLOCK_SIZE = 0x10
ACK_CMD = b"\x06"
MODES = ["NFM", "FM"]
SKIP_VALUES = ["S", ""]
TONES = chirp_common.TONES
DTCS = tuple(sorted(chirp_common.DTCS_CODES + (645,)))

# Special channels
SPECIALS = {
    "EMG": -2,
    "RLY": -1
    }

# Settings vars
TOT_LIST = ["Off"] + ["%s" % x for x in range(30, 210, 30)]
SCAN_TYPE_LIST = ["Time", "Carrier", "Search"]
LANGUAGE_LIST = ["Off", "English", "Chinese"]
TIMER_LIST = ["Off"] + ["%s h" % (x * 0.5) for x in range(1, 17)]
FM_RANGE_LIST = ["76-108", "65-76"]
RELAY_MODE_LIST = ["Off", "RX sync", "TX sync"]
BACKLIGHT_LIST = ["Off", "Key", "On"]
POWER_LIST = ["0.5 Watt", "1.0 Watt"]

# This is a general serial timeout for all serial read functions.
# Practice has show that about 0.07 sec will be enough to cover all radios.
STIMEOUT = 0.07

# this var controls the verbosity in the debug and by default it's low (False)
# make it True and you will to get a very verbose debug.log
debug = False

# #### ID strings #####################################################

# BF-T1 handheld
BFT1_magic = b"\x05PROGRAM"
BFT1_ident = b" BF9100S"


def _clean_buffer(radio):
    """Cleaning the read serial buffer, hard timeout to survive an infinite
    data stream"""

    dump = "1"
    datacount = 0

    try:
        while len(dump) > 0:
            dump = radio.pipe.read(100)
            datacount += len(dump)
            # hard limit to survive a infinite serial data stream
            # 5 times bigger than a normal rx block (20 bytes)
            if datacount > 101:
                seriale = "Please check your serial port selection."
                raise errors.RadioError(seriale)

    except Exception:
        raise errors.RadioError("Unknown error cleaning the serial buffer")


def _rawrecv(radio, amount=0):
    """Raw read from the radio device"""

    # var to hold the data to return
    data = b""

    try:
        if amount == 0:
            data = radio.pipe.read()
        else:
            data = radio.pipe.read(amount)

        # DEBUG
        if debug is True:
            LOG.debug("<== (%d) bytes:\n\n%s" %
                      (len(data), util.hexprint(data)))

        # fail if no data is received
        if len(data) == 0:
            raise errors.RadioError("No data received from radio")

    except:
        raise errors.RadioError("Error reading data from radio")

    return data


def _send(radio, data):
    """Send data to the radio device"""

    try:
        radio.pipe.write(data)

        # DEBUG
        if debug is True:
            LOG.debug("==> (%d) bytes:\n\n%s" %
                      (len(data), util.hexprint(data)))
    except:
        raise errors.RadioError("Error sending data to radio")


def _make_frame(cmd, addr, data=""):
    """Pack the info in the header format"""
    frame = struct.pack(">BHB", ord(cmd), addr, BLOCK_SIZE)

    # add the data if set
    if len(data) != 0:
        frame += data

    return frame


def _recv(radio, addr):
    """Get data from the radio"""

    # Get the full 20 bytes at a time
    # 4 bytes header + 16 bytes of data (BLOCK_SIZE)

    # get the whole block
    block = _rawrecv(radio, BLOCK_SIZE + 4)

    # short answer
    if len(block) < (BLOCK_SIZE + 4):
        raise errors.RadioError("Wrong block length (short) at 0x%04x" % addr)

    # long answer
    if len(block) > (BLOCK_SIZE + 4):
        raise errors.RadioError("Wrong block length (long) at 0x%04x" % addr)

    # header validation
    c, a, l = struct.unpack(">cHB", block[0:4])
    if c != b"W" or a != addr or l != BLOCK_SIZE:
        LOG.debug("Invalid header for block 0x%04x:" % addr)
        LOG.debug("CMD: %s  ADDR: %04x  SIZE: %02x" % (c, a, l))
        raise errors.RadioError("Invalid header for block 0x%04x:" % addr)

    # return the data, 16 bytes of payload
    return block[4:]


def _start_clone_mode(radio, status):
    """Put the radio in clone mode, 3 tries"""

    # cleaning the serial buffer
    _clean_buffer(radio)

    # prep the data to show in the UI
    status.cur = 0
    status.msg = "Identifying the radio..."
    status.max = 3
    radio.status_fn(status)

    try:
        for a in range(0, status.max):
            # Update the UI
            status.cur = a + 1
            radio.status_fn(status)

            # send the magic word
            _send(radio, radio._magic)

            # Now you get a x06 of ACK if all goes well
            ack = _rawrecv(radio, 1)

            if ack == ACK_CMD:
                # DEBUG
                LOG.info("Magic ACK received")
                status.cur = status.max
                radio.status_fn(status)

                return True

        return False

    except errors.RadioError:
        raise
    except Exception as e:
        raise errors.RadioError("Error sending Magic to radio:\n%s" % e)


def _do_ident(radio, status):
    """Put the radio in PROGRAM mode & identify it"""
    #  set the serial discipline (default)
    radio.pipe.baudrate = 9600
    radio.pipe.parity = "N"
    radio.pipe.bytesize = 8
    radio.pipe.stopbits = 1
    radio.pipe.timeout = STIMEOUT

    # open the radio into program mode
    if _start_clone_mode(radio, status) is False:
        raise errors.RadioError("Radio did not enter clone mode, wrong model?")

    # Ok, poke it to get the ident string
    _send(radio, b"\x02")
    ident = _rawrecv(radio, len(radio._id))

    # basic check for the ident
    if len(ident) != len(radio._id):
        raise errors.RadioError("Radio send a odd identification block.")

    # check if ident is OK
    if ident != radio._id:
        LOG.debug("Incorrect model ID, got this:\n\n" + util.hexprint(ident))
        raise errors.RadioError("Radio identification failed.")

    # handshake
    _send(radio, ACK_CMD)
    ack = _rawrecv(radio, 1)

    # checking handshake
    if len(ack) == 1 and ack == ACK_CMD:
        # DEBUG
        LOG.info("ID ACK received")
    else:
        LOG.debug("Radio handshake failed.")
        raise errors.RadioError("Radio handshake failed.")

    # DEBUG
    LOG.info("Positive ident, this is a %s %s" % (radio.VENDOR, radio.MODEL))

    return True


def _download(radio):
    """Get the memory map"""

    # UI progress
    status = chirp_common.Status()

    # put radio in program mode and identify it
    _do_ident(radio, status)

    # reset the progress bar in the UI
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning from radio..."
    status.cur = 0
    radio.status_fn(status)

    # cleaning the serial buffer
    _clean_buffer(radio)

    data = b""
    for addr in range(0, MEM_SIZE, BLOCK_SIZE):
        # sending the read request
        _send(radio, _make_frame(b"R", addr))

        # read
        d = _recv(radio, addr)

        # aggregate the data
        data += d

        # UI Update
        status.cur = addr // BLOCK_SIZE
        status.msg = "Cloning from radio..."
        radio.status_fn(status)

    # close comms with the radio
    _send(radio, b"\x62")
    # DEBUG
    LOG.info("Close comms cmd sent, radio must reboot now.")

    return data


def _upload(radio):
    """Upload procedure, we only upload to the radio the Writable space"""

    # UI progress
    status = chirp_common.Status()

    # put radio in program mode and identify it
    _do_ident(radio, status)

    # get the data to upload to radio
    data = radio.get_mmap()

    # Reset the UI progress
    status.max = WRITE_SIZE // BLOCK_SIZE
    status.cur = 0
    status.msg = "Cloning to radio..."
    radio.status_fn(status)

    # cleaning the serial buffer
    _clean_buffer(radio)

    # the fun start here, we use WRITE_SIZE instead of the full MEM_SIZE
    for addr in range(0, WRITE_SIZE, BLOCK_SIZE):
        # getting the block of data to send
        d = data[addr:addr + BLOCK_SIZE]

        # build the frame to send
        frame = _make_frame(b"W", addr, d)

        # send the frame
        _send(radio, frame)

        # receiving the response
        ack = _rawrecv(radio, 1)

        # basic check
        if len(ack) != 1:
            raise errors.RadioError("No ACK when writing block 0x%04x" % addr)

        if ack != ACK_CMD:
            raise errors.RadioError("Bad ACK writing block 0x%04x:" % addr)

        # UI Update
        status.cur = addr // BLOCK_SIZE
        status.msg = "Cloning to radio..."
        radio.status_fn(status)

    # close comms with the radio
    _send(radio, b"\x62")
    # DEBUG
    LOG.info("Close comms cmd sent, radio must reboot now.")


def _model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""

    # a reliable fingerprint: the model name at
    rid = data[0x06f8:0x0700]

    if rid == BFT1_ident:
        return True

    return False


def _decode_ranges(low, high):
    """Unpack the data in the ranges zones in the memmap and return
    a tuple with the integer corresponding to the MHz it means"""
    return (int(low) * 100000, int(high) * 100000)


MEM_FORMAT = """

struct channel {
  lbcd rxfreq[4];       // rx freq.
  u8 rxtone;            // x00 = none
                        // x01 - x32 = index of the analog tones
                        // x33 - x9b = index of Digital tones
                        // Digital tone polarity is handled below by
                        // ttondinv & ttondinv settings
  lbcd txoffset[4];     // the difference against RX, direction handled by
                        // offplus & offminus
  u8 txtone;            // Idem to rxtone
  u8 noskip:1,      // if true is included in the scan
     wide:1,        // 1 = Wide, 0 = Narrow
     ttondinv:1,    // if true TX tone is Digital & Inverted
     unA:1,         //
     rtondinv:1,    // if true RX tone is Digital & Inverted
     unB:1,         //
     offplus:1,     // TX = RX + offset
     offminus:1;    // TX = RX - offset
  u8 empty[5];
};

#seekto 0x0000;
struct channel emg;             // channel 0 is Emergent CH
#seekto 0x0010;
struct channel channels[20];    // normal 1-20 mem channels

#seekto 0x0150;     // Settings
struct {
  lbcd vhfl[2];     // VHF low limit
  lbcd vhfh[2];     // VHF high limit
  lbcd uhfl[2];     // UHF low limit
  lbcd uhfh[2];     // UHF high limit
  u8 unk0[8];
  u8 unk1[2];       // start of 0x0160 <=======
  u8 squelch;       // byte: 0-9
  u8 vox;           // byte: 0-9
  u8 timeout;       // tot, 0 off, then 30 sec increments up to 180
  u8 batsave:1,     // battery save 0 = off, 1 = on
     fm_funct:1,    // fm-radio 0=off, 1=on ( off disables fm button on set )
     ste:1,         // squelch tail 0 = off, 1 = on
     blo:1,         // busy lockout 0 = off, 1 = on
     beep:1,        // key beep 0 = off, 1 = on
     lock:1,        // keylock 0 = ff,  = on
     backlight:2;   // backlight 00 = off, 01 = key, 10 = on
  u8 scantype;      // scan type 0 = timed, 1 = carrier, 2 = stop
  u8 channel;       // active channel 1-20, setting it works on upload
  u8 fmrange;       // fm range 1 = low[65-76](ASIA), 0 = high[76-108](AMERICA)
  u8 alarm;         // alarm (count down timer)
                    //    d0 - d16 in half hour increments => off, 0.5 - 8.0 h
  u8 voice;         // voice prompt 0 = off, 1 = English, 2 = Chinese
  u8 volume;        // volume 1-7 as per the radio steps
                    //    set to #FF by original software on upload
                    //    chirp uploads actual value and works.
  u16 fm_vfo;       // the frequency of the fm receiver.
                    //    resulting frequency is 65 + value * 0.1 MHz
                    //    0x145 is then 65 + 325*0.1 = 97.5 MHz
  u8 relaym;        // relay mode, d0 = off, d2 = re-tx, d1 = re-rx
                    //    still a mystery on how it works
  u8 tx_pwr;        // tx pwr 0 = low (0.5W), 1 = high(1.0W)
} settings;

#seekto 0x0170;     // Relay CH
struct channel rly;

"""


@directory.register
class BFT1(chirp_common.CloneModeRadio, chirp_common.ExperimentalRadio):
    """Baofeng BT-F1 radio & possibly alike radios"""
    VENDOR = "Baofeng"
    MODEL = "BF-T1"
    NEEDS_COMPAT_SERIAL = False
    _vhf_range = (130000000, 174000000)
    _uhf_range = (400000000, 520000000)
    _upper = 20
    _magic = BFT1_magic
    _id = BFT1_ident
    _bw_shift = False

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This driver is experimental.\n'
             '\n'
             'Please keep a copy of your memories with the original software '
             'if you treasure them, this driver is new and may contain'
             ' bugs.\n'
             '\n'
             '"Emergent CH" & "Relay CH" are implemented via special channels,'
             'be sure to click on the button on the interface to access them.'
             )
        rp.pre_download = _(
            "Follow these instructions to download your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow these instructions to upload your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def get_features(self):
        """Get the radio's features"""

        rf = chirp_common.RadioFeatures()
        rf.valid_special_chans = list(SPECIALS.keys())
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = False
        rf.has_offset = True
        rf.has_mode = True
        rf.valid_modes = MODES
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.valid_duplexes = ["", "-", "+", "split"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_skips = SKIP_VALUES
        rf.valid_dtcs_codes = DTCS
        rf.memory_bounds = (1, self._upper)
        rf.valid_tuning_steps = [2.5, 5., 6.25, 10., 12.5, 25.]

        # normal dual bands
        rf.valid_bands = [self._vhf_range, self._uhf_range]

        return rf

    def process_mmap(self):
        """Process the mem map into the mem object"""

        # Get it
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

        # set the band limits as the memmap
        settings = self._memobj.settings
        self._vhf_range = _decode_ranges(settings.vhfl, settings.vhfh)
        self._uhf_range = _decode_ranges(settings.uhfl, settings.uhfh)

    def sync_in(self):
        """Download from radio"""
        data = _download(self)
        self._mmap = memmap.MemoryMapBytes(data)
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""

        try:
            _upload(self)
        except errors.RadioError:
            raise
        except Exception as e:
            raise errors.RadioError("Error: %s" % e)

    def _decode_tone(self, val, inv):
        """Parse the tone data to decode from mem, it returns:
        Mode (''|DTCS|Tone), Value (None|###), Polarity (None,N,R)"""

        if val == 0:
            return '', None, None
        elif val < 51:  # analog tone
            return 'Tone', TONES[val - 1], None
        elif val > 50:  # digital tone
            pol = "N"
            # polarity?
            if inv == 1:
                pol = "R"

            return 'DTCS', DTCS[val - 51], pol

    def _encode_tone(self, memtone, meminv, mode, tone, pol):
        """Parse the tone data to encode from UI to mem"""

        if mode == '' or mode is None:
            memtone.set_value(0)
            meminv.set_value(0)
        elif mode == 'Tone':
            # caching errors for analog tones.
            try:
                memtone.set_value(TONES.index(tone) + 1)
                meminv.set_value(0)
            except:
                msg = "TCSS Tone '%d' is not supported" % tone
                LOG.error(msg)
                raise errors.RadioError(msg)

        elif mode == 'DTCS':
            # caching errors for digital tones.
            try:
                memtone.set_value(DTCS.index(tone) + 51)
                if pol == "R":
                    meminv.set_value(True)
                else:
                    meminv.set_value(False)
            except:
                msg = "Digital Tone '%d' is not supported" % tone
                LOG.error(msg)
                raise errors.RadioError(msg)
        else:
            msg = "Internal error: invalid mode '%s'" % mode
            LOG.error(msg)
            raise errors.InvalidDataError(msg)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number])

    def _get_special(self, number):
        if isinstance(number, str):
            return (getattr(self._memobj, number.lower()))
        elif number < 0:
            for k, v in SPECIALS.items():
                if number == v:
                    return (getattr(self._memobj, k.lower()))
        else:
            return self._memobj.channels[number-1]

    def get_memory(self, number):
        """Get the mem representation from the radio image"""
        _mem = self._get_special(number)

        # Create a high-level memory object to return to the UI
        mem = chirp_common.Memory()

        # Check if special or normal
        if isinstance(number, str):
            mem.number = SPECIALS[number]
            mem.extd_number = number
        else:
            mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xFF":
            mem.empty = True
            return mem

        # Freq and offset
        mem.freq = int(_mem.rxfreq) * 10

        # TX freq (Stored as a difference)
        mem.offset = int(_mem.txoffset) * 10
        mem.duplex = ""

        # must work out the polarity
        if mem.offset != 0:
            if _mem.offminus == 1:
                mem.duplex = "-"
                #  tx below RX

            if _mem.offplus == 1:
                #  tx above RX
                mem.duplex = "+"

            # split RX/TX in different bands
            if mem.offset > 71000000:
                mem.duplex = "split"

                # show the actual value in the offset, depending on the shift
                if _mem.offminus == 1:
                    mem.offset = mem.freq - mem.offset
                if _mem.offplus == 1:
                    mem.offset = mem.freq + mem.offset

        # wide/narrow
        mem.mode = MODES[int(_mem.wide)]

        # skip
        mem.skip = SKIP_VALUES[_mem.noskip]

        # tone data
        rxtone = txtone = None
        txtone = self._decode_tone(_mem.txtone, _mem.ttondinv)
        rxtone = self._decode_tone(_mem.rxtone, _mem.rtondinv)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

        return mem

    def set_memory(self, mem):
        """Set the memory data in the eeprom img from the UI"""
        # get the eprom representation of this channel
        _mem = self._get_special(mem.number)

        # if empty memory
        if mem.empty:
            # the channel itself
            _mem.set_raw("\xFF" * 16)
            # return it
            return mem

        # frequency
        _mem.rxfreq = mem.freq / 10

        # duplex/ offset Offset is an absolute value
        _mem.txoffset = mem.offset / 10

        # must work out the polarity
        if mem.duplex == "":
            _mem.offplus = 0
            _mem.offminus = 0
        elif mem.duplex == "+":
            _mem.offplus = 1
            _mem.offminus = 0
        elif mem.duplex == "-":
            _mem.offplus = 0
            _mem.offminus = 1
        elif mem.duplex == "split":
            if mem.freq > mem.offset:
                _mem.offplus = 0
                _mem.offminus = 1
                _mem.txoffset = (mem.freq - mem.offset) / 10
            else:
                _mem.offplus = 1
                _mem.offminus = 0
                _mem.txoffset = (mem.offset - mem.freq) / 10

        # wide/narrow
        _mem.wide = MODES.index(mem.mode)

        # skip
        _mem.noskip = SKIP_VALUES.index(mem.skip)

        # tone data
        ((txmode, txtone, txpol), (rxmode, rxtone, rxpol)) = \
            chirp_common.split_tone_encode(mem)
        self._encode_tone(_mem.txtone, _mem.ttondinv, txmode, txtone, txpol)
        self._encode_tone(_mem.rxtone, _mem.rtondinv, rxmode, rxtone, rxpol)

        return mem

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        fm = RadioSettingGroup("fm", "FM Radio")
        adv = RadioSettingGroup("adv", "Advanced Settings")
        group = RadioSettings(basic, fm, adv)

        # ## Basic Settings
        rs = RadioSetting("tx_pwr", "TX Power",
                          RadioSettingValueList(
                            POWER_LIST, POWER_LIST[_settings.tx_pwr]))
        basic.append(rs)

        rs = RadioSetting("channel", "Active Channel",
                          RadioSettingValueInteger(1, 20, _settings.channel))
        basic.append(rs)

        rs = RadioSetting("squelch", "Squelch Level",
                          RadioSettingValueInteger(0, 9, _settings.squelch))
        basic.append(rs)

        rs = RadioSetting("vox", "VOX Level",
                          RadioSettingValueInteger(0, 9, _settings.vox))
        basic.append(rs)

        # volume validation, as the OEM software set 0xFF on write
        _volume = _settings.volume
        if _volume > 7:
            _volume = 7
        rs = RadioSetting("volume", "Volume Level",
                          RadioSettingValueInteger(0, 7, _volume))
        basic.append(rs)

        rs = RadioSetting("scantype", "Scan Type",
                          RadioSettingValueList(SCAN_TYPE_LIST, SCAN_TYPE_LIST[
                              _settings.scantype]))
        basic.append(rs)

        rs = RadioSetting("timeout", "Time Out Timer (seconds)",
                          RadioSettingValueList(
                            TOT_LIST, TOT_LIST[_settings.timeout]))
        basic.append(rs)

        rs = RadioSetting("voice", "Voice Prompt",
                          RadioSettingValueList(
                            LANGUAGE_LIST, LANGUAGE_LIST[_settings.voice]))
        basic.append(rs)

        rs = RadioSetting("alarm", "Alarm Time",
                          RadioSettingValueList(
                            TIMER_LIST, TIMER_LIST[_settings.alarm]))
        basic.append(rs)

        rs = RadioSetting("backlight", "Backlight",
                          RadioSettingValueList(
                            BACKLIGHT_LIST,
                            BACKLIGHT_LIST[_settings.backlight]))
        basic.append(rs)

        rs = RadioSetting("blo", "Busy Lockout",
                          RadioSettingValueBoolean(_settings.blo))
        basic.append(rs)

        rs = RadioSetting("ste", "Squelch Tail Eliminate",
                          RadioSettingValueBoolean(_settings.ste))
        basic.append(rs)

        rs = RadioSetting("batsave", "Battery Save",
                          RadioSettingValueBoolean(_settings.batsave))
        basic.append(rs)

        rs = RadioSetting("lock", "Key Lock",
                          RadioSettingValueBoolean(_settings.lock))
        basic.append(rs)

        rs = RadioSetting("beep", "Key Beep",
                          RadioSettingValueBoolean(_settings.beep))
        basic.append(rs)

        # ## FM Settings
        rs = RadioSetting("fm_funct", "FM Function",
                          RadioSettingValueBoolean(_settings.fm_funct))
        fm.append(rs)

        rs = RadioSetting("fmrange", "FM Range",
                          RadioSettingValueList(
                            FM_RANGE_LIST, FM_RANGE_LIST[_settings.fmrange]))
        fm.append(rs)

        # callbacks for the FM VFO
        def apply_fm_freq(setting, obj):
            value = int(setting.value.get_value() * 10) - 650
            LOG.debug("Setting fm_vfo = %s" % (value))
            if self._bw_shift:
                value = ((value & 0x00FF) << 8) | ((value & 0xFF00) >> 8)
            setattr(obj, setting.get_name(), value)

        # broadcast FM setting
        value = _settings.fm_vfo
        value_shifted = ((value & 0x00FF) << 8) | ((value & 0xFF00) >> 8)
        if value_shifted <= 108.0 * 10 - 650:
            # storage method 3 (discovered 2022)
            self._bw_shift = True
            _fm_vfo = value_shifted / 10.0 + 65
        elif value <= 108.0 * 10 - 650:
            # original storage method (2012)
            _fm_vfo = value / 10.0 + 65
        else:
            # unknown (undiscovered method or no FM chip?)
            _fm_vfo = False
        if _fm_vfo:
            rs = RadioSetting("fm_vfo", "FM Station",
                              RadioSettingValueFloat(65, 108, _fm_vfo))
            rs.set_apply_callback(apply_fm_freq, _settings)
            fm.append(rs)

        # ## Advanced
        def apply_limit(setting, obj):
            setattr(obj, setting.get_name(), int(setting.value) * 10)

        rs = RadioSetting("vhfl", "VHF Low Limit",
                          RadioSettingValueInteger(130, 174, int(
                              _settings.vhfl) / 10))
        rs.set_apply_callback(apply_limit, _settings)
        adv.append(rs)

        rs = RadioSetting("vhfh", "VHF High Limit",
                          RadioSettingValueInteger(130, 174, int(
                              _settings.vhfh) / 10))
        rs.set_apply_callback(apply_limit, _settings)
        adv.append(rs)

        rs = RadioSetting("uhfl", "UHF Low Limit",
                          RadioSettingValueInteger(400, 520, int(
                              _settings.uhfl) / 10))
        rs.set_apply_callback(apply_limit, _settings)
        adv.append(rs)

        rs = RadioSetting("uhfh", "UHF High Limit",
                          RadioSettingValueInteger(400, 520, int(
                              _settings.uhfh) / 10))
        rs.set_apply_callback(apply_limit, _settings)
        adv.append(rs)

        rs = RadioSetting("relaym", "Relay Mode",
                          RadioSettingValueList(RELAY_MODE_LIST,
                                                RELAY_MODE_LIST[_settings.relaym]))
        adv.append(rs)

        return group

    def set_settings(self, uisettings):
        _settings = self._memobj.settings

        for element in uisettings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                name = element.get_name()
                value = element.value

                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    obj = getattr(_settings, name)
                    setattr(_settings, name, value)

                LOG.debug("Setting %s: %s" % (name, value))
            except Exception:
                LOG.debug(element.get_name())
                raise

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # testing the file data size
        if len(filedata) == MEM_SIZE:
            match_size = True

            # DEBUG
            if debug is True:
                LOG.debug("BF-T1 matched!")

        # testing the firmware model fingerprint
        match_model = _model_match(cls, filedata)

        return match_size and match_model



# Copyright 2013 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import bandplan, bandplan_iaru_r2


SHORTNAME = "north_america"

DESC = {
  "name": "North American Band Plan",
  "url":  "http://www.arrl.org/band-plan"
}

BANDS_160M = (
  bandplan.Band((1800000, 2000000), "160 Meter Band", mode="CW"),
  bandplan.Band((1800000, 1810000), "Digital Modes"),
  bandplan.Band((1843000, 2000000), "SSB, SSTV and other wideband modes"),
  bandplan.Band((1995000, 2000000), "Experimental"),
  bandplan.Band((1999000, 2000000), "Beacons"),
)

BANDS_80M = (
  bandplan.Band((3500000, 4000000), "80 Meter Band"),
  bandplan.Band((3570000, 3600000), "RTTY/Data", mode="RTTY"),
  bandplan.Band((3790000, 3800000), "DX window"),
)

BANDS_40M = (
  bandplan.Band((7000000, 7300000), "40 Meter Band"),
  bandplan.Band((7080000, 7125000), "RTTY/Data", mode="RTTY"),
)

BANDS_30M = (
  bandplan.Band((10100000, 10150000), "30 Meter Band"),
  bandplan.Band((10130000, 10140000), "RTTY", mode="RTTY"),
  bandplan.Band((10140000, 10150000), "Packet"),
)

BANDS_20M = (
  bandplan.Band((14000000, 14350000), "20 Meter Band"),
  bandplan.Band((14070000, 14095000), "RTTY", mode="RTTY"),
  bandplan.Band((14095000, 14099500), "Packet"),
  bandplan.Band((14100500, 14112000), "Packet"),
)

BANDS_17M = (
  bandplan.Band((18068000, 18168000), "17 Meter Band"),
  bandplan.Band((18100000, 18105000), "RTTY", mode="RTTY"),
  bandplan.Band((18105000, 18110000), "Packet"),
)

BANDS_15M = (
  bandplan.Band((21000000, 21450000), "15 Meter Band"),
  bandplan.Band((21070000, 21110000), "RTTY/Data", mode="RTTY"),
)

BANDS_12M = (
  bandplan.Band((24890000, 24990000), "12 Meter Band"),
  bandplan.Band((24920000, 24925000), "RTTY", mode="RTTY"),
  bandplan.Band((24925000, 24930000), "Packet"),
)

BANDS_10M = (
  bandplan.Band((28000000, 29700000), "10 Meter Band"),
  bandplan.Band((28000000, 28070000), "CW", mode="CW"),
  bandplan.Band((28070000, 28150000), "RTTY", mode="RTTY"),
  bandplan.Band((28150000, 28190000), "CW", mode="CW"),
  bandplan.Band((28201000, 28300000), "Beacons", mode="CW"),
  bandplan.Band((28300000, 29300000), "Phone"),
  bandplan.Band((29000000, 29200000), "AM", mode="AM"),
  bandplan.Band((29300000, 29510000), "Satellite Downlinks"),
  bandplan.Band((29520000, 29590000), "Repeater Inputs",
                step_khz=10, mode="FM"),
  bandplan.Band((29610000, 29700000), "Repeater Outputs",
                step_khz=10, mode="FM", input_offset=-890000),
)

BANDS_6M = (
  bandplan.Band((50000000, 54000000), "6 Meter Band"),
  bandplan.Band((50000000, 50100000), "CW, beacons", mode="CW"),
  bandplan.Band((50060000, 50080000), "beacon subband"),
  bandplan.Band((50100000, 50300000), "SSB, CW", mode="USB"),
  bandplan.Band((50100000, 50125000), "DX window", mode="USB"),
  bandplan.Band((50300000, 50600000), "All modes"),
  bandplan.Band((50600000, 50800000), "Nonvoice communications"),
  bandplan.Band((50800000, 51000000), "Radio remote control", step_khz=20),
  bandplan.Band((51000000, 51100000), "Pacific DX window"),
  bandplan.Band((51120000, 51180000), "Digital repeater inputs", step_khz=10),
  bandplan.Band((51500000, 51600000), "Simplex"),
  bandplan.Band((51620000, 51980000), "Repeater outputs A",
                input_offset=-500000),
  bandplan.Band((51620000, 51680000), "Digital repeater outputs",
                input_offset=-500000),
  bandplan.Band((52020000, 52040000), "FM simplex", mode="FM"),
  bandplan.Band((52500000, 52980000), "Repeater outputs B",
                input_offset=-500000, step_khz=20, mode="FM"),
  bandplan.Band((53000000, 53100000), "FM simplex", mode="FM"),
  bandplan.Band((53100000, 53400000), "Radio remote control", step_khz=100),
  bandplan.Band((53500000, 53980000), "Repeater outputs C",
                input_offset=-500000),
  bandplan.Band((53500000, 53800000), "Radio remote control", step_khz=100),
  bandplan.Band((53520000, 53900000), "Simplex"),
)

BANDS_2M = (
  bandplan.Band((144000000, 148000000), "2 Meter Band"),
  bandplan.Band((144000000, 144050000), "EME (CW)", mode="CW"),
  bandplan.Band((144050000, 144100000), "General CW and weak signals",
                mode="CW"),
  bandplan.Band((144100000, 144200000), "EME and weak-signal SSB",
                mode="USB"),
  bandplan.Band((144200000, 144275000), "General SSB operation",
                mode="USB"),
  bandplan.Band((144275000, 144300000), "Propagation beacons", mode="CW"),
  bandplan.Band((144300000, 144500000), "OSCAR subband"),
  bandplan.Band((144600000, 144900000), "FM repeater inputs", mode="FM"),
  bandplan.Band((144900000, 145100000), "Weak signal and FM simplex",
                mode="FM", step_khz=10),
  bandplan.Band((145100000, 145200000), "Linear translator outputs",
                input_offset=-600000),
  bandplan.Band((145200000, 145500000), "FM repeater outputs",
                input_offset=-600000, mode="FM",),
  bandplan.Band((145500000, 145800000), "Misc and experimental modes"),
  bandplan.Band((145800000, 146000000), "OSCAR subband"),
  bandplan.Band((146400000, 146580000), "Simplex"),
  bandplan.Band((146610000, 146970000), "Repeater outputs",
                input_offset=-600000),
  bandplan.Band((147000000, 147390000), "Repeater outputs",
                input_offset=600000),
  bandplan.Band((147420000, 147570000), "Simplex"),
)

BANDS_1_25M = (
  bandplan.Band((222000000, 225000000), "1.25 Meter Band"),
  bandplan.Band((222000000, 222150000), "Weak-signal modes"),
  bandplan.Band((222000000, 222025000), "EME"),
  bandplan.Band((222050000, 222060000), "Propagation beacons"),
  bandplan.Band((222100000, 222150000), "Weak-signal CW & SSB"),
  bandplan.Band((222150000, 222250000), "Local coordinator's option"),
  bandplan.Band((223400000, 223520000), "FM simplex", mode="FM"),
  bandplan.Band((223520000, 223640000), "Digital, packet"),
  bandplan.Band((223640000, 223700000), "Links, control"),
  bandplan.Band((223710000, 223850000), "Local coordinator's option"),
  bandplan.Band((223850000, 224980000), "Repeater outputs only",
                mode="FM", input_offset=-1600000),
)

BANDS_70CM = (
  bandplan.Band((420000000, 450000000), "70 Centimeter Band"),
  bandplan.Band((420000000, 426000000), "ATV repeater or simplex"),
  bandplan.Band((426000000, 432000000), "ATV simplex"),
  bandplan.Band((432000000, 432070000), "EME (Earth-Moon-Earth)"),
  bandplan.Band((432070000, 432100000), "Weak-signal CW", mode="CW"),
  bandplan.Band((432100000, 432300000), "Mixed-mode and weak-signal work"),
  bandplan.Band((432300000, 432400000), "Propagation beacons"),
  bandplan.Band((432400000, 433000000), "Mixed-mode and weak-signal work"),
  bandplan.Band((433000000, 435000000), "Auxiliary/repeater links"),
  bandplan.Band((435000000, 438000000), "Satellite only (internationally)"),
  bandplan.Band((438000000, 444000000), "ATV repeater input/repeater links",
                input_offset=5000000),
  bandplan.Band((442000000, 445000000), "Repeater input/output (local option)",
                input_offset=5000000),
  bandplan.Band((445000000, 447000000), "Shared by aux and control links, "
                "repeaters, simplex (local option)"),
  bandplan.Band((447000000, 450000000), "Repeater inputs and outputs "
                "(local option)", input_offset=-5000000),
)

BANDS_33CM = (
  bandplan.Band((902000000, 928000000), "33 Centimeter Band"),
  bandplan.Band((902075000, 902100000), "CW/SSB, Weak signal"),
  bandplan.Band((902100000, 902125000), "CW/SSB, Weak signal"),
  bandplan.Band((903000000, 903100000), "CW/SSB, Beacons and weak signal"),
  bandplan.Band((903100000, 903400000), "CW/SSB, Weak signal"),
  bandplan.Band((903400000, 909000000), "Mixed modes, Mixed operations "
                "including control links"),
  bandplan.Band((909000000, 915000000), "Analog/digital Broadband multimedia "
                "including ATV, DATV and SS"),
  bandplan.Band((915000000, 921000000), "Analog/digital Broadband multimedia "
                "including ATV, DATV and SS"),
  bandplan.Band((921000000, 927000000), "Analog/digital Broadband multimedia "
                "including ATV, DATV and SS"),
  bandplan.Band((927000000, 927075000), "FM / other including DV or CW/SSB",
                input_offset=-25000000, step_khz=12.5),
  bandplan.Band((927075000, 927125000), "FM / other including DV. Simplex"),
  bandplan.Band((927125000, 928000000), "FM / other including DV",
                input_offset=-25000000, step_khz=12.5),
)

BANDS_23CM = (
  bandplan.Band((1240000000, 1300000000), "23 Centimeter Band"),
  bandplan.Band((1240000000, 1246000000), "ATV Channel #1"),
  bandplan.Band((1246000000, 1248000000), "Point-to-point links paired "
                "with 1258.000-1260.000", mode="FM"),
  bandplan.Band((1248000000, 1252000000), "Digital"),
  bandplan.Band((1252000000, 1258000000), "ATV Channel #2"),
  bandplan.Band((1258000000, 1260000000),
                "Point-to-point links paired with 1246.000-1248.000",
                mode="FM"),
  bandplan.Band((1240000000, 1260000000), "Regional option, FM ATV"),
  bandplan.Band((1260000000, 1270000000), "Satellite uplinks, Experimental, "
                "Simplex ATV"),
  bandplan.Band((1270000000, 1276000000), "FM, digital Repeater inputs "
                "(Regional option)", step_khz=25),
  bandplan.Band((1276000000, 1282000000), "ATV Channel #3"),
  bandplan.Band((1282000000, 1288000000), "FM, digital repeater outputs",
                step_khz=25, input_offset=-12000000),
  bandplan.Band((1288000000, 1294000000), "Various Broadband Experimental, "
                "Simplex ATV"),
  bandplan.Band((1290000000, 1294000000), "FM, digital Repeater outputs "
                "(Regional option)", step_khz=25, input_offset=-20000000),
  bandplan.Band((1294000000, 1295000000), "FM simplex", mode="FM"),
  bandplan.Band((1295000000, 1297000000), "Narrow Band Segment"),
  bandplan.Band((1295000000, 1295800000), "Narrow Band Image, Experimental"),
  bandplan.Band((1295800000, 1296080000), "CW, SSB, digital EME"),
  bandplan.Band((1296080000, 1296200000), "CW, SSB Weak Signal"),
  bandplan.Band((1296200000, 1296400000), "CW, digital Beacons"),
  bandplan.Band((1296400000, 1297000000), "General Narrow Band"),
  bandplan.Band((1297000000, 1300000000), "Digital"),
)

BANDS_13CM = (
  bandplan.Band((2300000000, 2450000000), "13 Centimeter Band"),
  bandplan.Band((2300000000, 2303000000), "Analog & Digital 0.05-1.0 MHz, "
                "including full duplex; paired with 2390-2393"),
  bandplan.Band((2303000000, 2303750000), "Analog & Digital <50 kHz; "
                "paired with 2393 - 2393.750"),
  bandplan.Band((2303750000, 2304000000), "SSB, CW, digital weak-signal"),
  bandplan.Band((2304000000, 2304100000), "Weak Signal EME Band, <3 kHz"),
  bandplan.Band((2304100000, 2304300000),
                "SSB, CW, digital weak-signal, <3 kHz"),
  bandplan.Band((2304300000, 2304400000), "Beacons, <3 kHz"),
  bandplan.Band((2304400000, 2304750000), "SSB, CW, digital weak-signal and "
                "NBFM, <6 kHz"),
  bandplan.Band((2304750000, 2305000000), "Analog & Digital; paired with "
                "2394.750-2395, <50 kHz"),
  bandplan.Band((2305000000, 2310000000), "Analog & Digital, paired with "
                "2395-2400, 0.05 - 1.0 MHz"),
  bandplan.Band((2310000000, 2390000000), "NON-AMATEUR"),
  bandplan.Band((2390000000, 2393000000), "Analog & Digital, including full "
                "duplex; paired with 2300-2303, 0.05 - 1.0 MHz"),
  bandplan.Band((2393000000, 2393750000), "Analog & Digital; paired with "
                "2303-2303.750, < 50 kHz"),
  bandplan.Band((2393750000, 2394750000), "Experimental"),
  bandplan.Band((2394750000, 2395000000), "Analog & Digital; paired with "
                "2304.750-2305, < 50 kHz"),
  bandplan.Band((2395000000, 2400000000), "Analog & Digital, including full "
                "duplex; paired with 2305-2310, 0.05-1.0 MHz"),
  bandplan.Band((2400000000, 2410000000), "Amateur Satellite Communications, "
                "<6 kHz"),
  bandplan.Band((2410000000, 2450000000), "Broadband Modes, 22 MHz max."),
)

BANDS = bandplan_iaru_r2.BANDS + \
        BANDS_160M + BANDS_80M + BANDS_40M + BANDS_30M + BANDS_20M + \
        BANDS_17M + BANDS_15M + BANDS_12M + BANDS_10M + BANDS_6M + \
        BANDS_2M + BANDS_1_25M + BANDS_70CM + BANDS_33CM + BANDS_23CM + \
        BANDS_13CM


# US 2017+ GMRS/FRS Frequencies
GMRS_LOW = [462562500, 462587500, 462612500, 462637500, 462662500,
            462687500, 462712500]
GMRS_HHONLY = [467562500, 467587500, 467612500, 467637500, 467662500,
               467687500, 467712500]
GMRS_HIRPT = [462550000, 462575000, 462600000, 462625000, 462650000,
              462675000, 462700000, 462725000]
ALL_GMRS_FREQS = GMRS_LOW + GMRS_HHONLY + GMRS_HIRPT


# US MURS Frequencies
MURS_NFM = [151820000, 151880000, 151940000]
MURS_FM = [154570000, 154600000]
ALL_MURS_FREQS = MURS_NFM + MURS_FM



# -*- coding: utf-8 -*-
# Copyright 2018-2020 Jaroslav Škarvada <jskarvad@redhat.com>
# Based on various code from the CHIRP

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import icf
from chirp import chirp_common, bitwise, errors, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettings
import argparse

ICX90_MEM_FORMAT = """
struct mem_item {
  ul24 freq;
  u8 dtcs_polarity:2,
     unknown_1:2,
     offset_freq_mult:1,
     unknown_2:2,
     freq_mult:1;
  u8 unknown_3:1,
     duplex:2,
     mode:2,
     tone_mode:3;
  ul16 offset_freq;
  u8 dtcs;
  u8 tx_tone_lo:4,
     tune_step:4;
  u8 rx_tone:6,
     tx_tone_hi:2;
  char name[6];
};
struct bank_item {
  u8 invisible_channel:1,
     prog_skip:1,
     mem_skip:1,
     bank_index:5;
  u8 bank_channel;
};
struct tv_mem_item {
  u8 fixed:7,
     mode:1;
  ul24 freq;
  char name[4];
};

struct mem_item memory[500];
struct mem_item scan_edges[50];
struct bank_item banks[500];
u8 unknown_4[120];
struct mem_item vfo_a_band[10];
struct mem_item vfo_b_band[10];
struct mem_item call_channels[5];
struct tv_mem_item tv_memory[68];
u8 unknown_5[35];
ul16 mem_channel;
u8 autodial;
u8 unknown_6[8];
u8 unknown_7:6,
   skip_scan:1,
   unknown_8:1;
u8 squelch_level;
struct {
  char dtmf_digits[16];
} dtmf_codes[10];
u8 tv_channel_skip[68];
u8 unknown_9[128];
u8 scan_resume;
u8 scan_pause;
u8 unknown_10;
u8 beep_volume;
u8 beep;
u8 backlight;
u8 busy_led;
u8 auto_power_off;
u8 power_save;
u8 monitor;
u8 dial_speedup;
u8 unknown_11;
u8 auto_repeater;
u8 dtmf_speed;
u8 hm_75a_function;
u8 wx_alert;
u8 expand_1;
u8 scan_stop_beep;
u8 scan_stop_light;
u8 unknown_12;
u8 light_position;
u8 light_color;
u8 unknown_13;
u8 band_edge_beep;
u8 auto_power_on;
u8 key_lock;
u8 ptt_lock;
u8 lcd_contrast;
u8 opening_message;
u8 expand_2;
u8 unknown_14;
u8 busy_lockout;
u8 timeout_timer;
u8 active_band;
u8 split;
u8 fm_narrow;
u8 morse_code_enable;
u8 morse_code_speed;
u8 unknown_15[22];
char opening_message_text[6];
u8 unknown_16[186];
u8 unknown_17:4,
   tune_step:4;
u8 unknown_18[4];
u8 band_selected;
u8 unknown_19[2];
u8 unknown_20:2,
   attenuator:1,
   vfo:1,
   power:1,
   dial_select:1,
   memory_name:1,
   memory_display:1;
u8 unknown_21[2];
u8 mode:4,
   unknown_22:4;
u8 unknown_23[9];
char alpha_tag[6];
u8 vfo_scan;
u8 memory_scan;
u8 unknown_24;
u8 tv_channel;
u8 wx_channel;
char comment[16];
"""

LOG = logging.getLogger(__name__)

# in bytes
MEM_ITEM_SIZE = 16
TV_MEM_ITEM_SIZE = 8

BANK_INDEX = ["A", "B", "C", "D", "E", "F", "G", "H", "J", "L", "N", "O",
              "P", "Q", "R", "T", "U", "Y"]
MEM_NUM = 500
BANKS = 18
BANK_NUM = 100
BANK_INDEX_NUM = len(BANK_INDEX)
DTMF_AUTODIAL_NUM = 10
DTMF_DIGITS_NUM = 16
OPENING_MESSAGE_LEN = 6
COMMENT_LEN = 16
BANDS = 10
TV_CHANNELS = 68

CHARSET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789()*+-,/|= "
NAME_LENGTH = 6
TV_NAME_LENGTH = 4
DUPLEX = ["", "-", "+", ""]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]
TONE_MODE = ["", "Tone", "TSQL", "DTCS"]
TUNE_STEP = [5.0, 6.25, 8.33, 9.0, 10.0, 12.5, 15.0, 20.0, 25.0, 30.0, 50.0, 100.0, 200.0]
TUNE_STEP_STR = [str(x) for x in TUNE_STEP]
MODE = ["FM", "WFM", "AM"]
TV_MODE = ["WFM", "AM"]

SQUELCH_LEVEL = ["Open", "Auto", "Level 1", "Level 2", "Level 3", "Level 4", "Level 5",
                 "Level 6", "Level 7", "Level 8", "Level 9"]
AUTODIAL = ["Tone call", "D0", "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9"]
DTMF_SPEED = ["100 ms", "200 ms", "300 ms", "400 ms"]
SCAN_RESUME = ["0 s", "1 s", "2 s", "3 s", "4 s", "5 s", "Hold"]
SCAN_PAUSE = ["2 s", "4 s", "6 s", "8 s", "10 s", "12 s", "14 s", "16 s", "18 s", "20 s", "Hold"]
BEEP_VOLUME = ["Volume", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12",
               "13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "23", "24", "25",
               "26", "27", "28", "29", "30", "31"]
BACKLIGHT = ["Off", "On", "Auto"]
AUTO_POWER_OFF = ["Off", "30 min", "60 min", "90 min", "120 min"]
POWER_SAVE = ["Off", "1:1", "1:4", "1:8", "1:16", "Auto"]
MONITOR = ["Push", "Hold"]
AUTO_REPEATER = ["Off", "Duplex only", "Duplex & tone"]
HM_75A_FUNCTION = ["Simple", "Normal 1", "Normal 2"]
LIGHT_POSITION = ["LCD", "Key", "All"]
LIGHT_COLOR = ["Green", "Orange", "Red"]
AUTO_POWER_ON = ["Off", "00:30", "01:00", "01:30", "02:00", "02:30", "03:00", "03:30", "04:00",
                 "04:30", "05:00", "05:30", "06:00", "06:30", "07:00", "07:30", "08:00",
                 "08:30", "09:00", "09:30", "10:00", "10:30", "11:00", "11:30", "12:00",
                 "12:30", "13:00", "13:30", "14:00", "14:30", "15:00", "15:30", "16:00",
                 "16:30", "17:00", "17:30", "18:00", "18:30", "19:00", "19:30", "20:00",
                 "20:30", "21:00", "21:30", "22:00", "22:30", "23:00", "23:30", "24:00"]
KEY_LOCK = ["Normal", "No squelch", "No volume", "All"]
LCD_CONTRAST = ["1", "2", "3", "4"]
TIMEOUT_TIMER = ["Off", "1 min", "3 min", "5 min", "10 min"]
ACTIVE_BAND = ["All", "Single"]
MORSE_CODE_SPEED = ["10 WPM", "15 WPM", "20 WPM", "25 WPM"]
WX_CHANNEL = ["WX01", "WX02", "WX03", "WX04", "WX05", "WX06", "WX07", "WX08", "WX09", "WX10"]
MEMORY_DISPLAY = ["Channel", "Bank"]
DIAL_SELECT = ["Normal", "Volume"]
POWER = ["High", "Low"]
VFO = ["A", "B"]
OPERATION_MODE = ["VFO", "Memory", "Call channel", "TV"]
VFO_SCAN = ["All", "Band", "P0", "P1", "P2", "P3", "P4", "P5", "P6", "P7", "P8",
            "P9", "P10", "P11", "P12", "P13", "P14", "P15", "P16", "P17", "P18",
            "P19", "P20", "P21", "P22", "P23", "P24"]
MEMORY_SCAN = ["All", "Bank", "Sel BC", "Sel 5 MHz", "Sel 50 MHz", "Sel WFM", "Sel Air",
               "Sel 144 MHz", "Sel 220 MHz", "Sel 300 MHz", "Sel 440 MHz", "Sel 800 MHz"]


class ICx90BankModel(icf.IcomIndexedBankModel):
    bank_index = BANK_INDEX

    def get_mappings(self):
        banks = []

        if (self._radio._num_banks != len(type(self).bank_index)):
            raise Exception("Invalid number of banks %d, supported only %d banks" %
                            (self._radio._num_banks, len(type(self).bank_index)))

        for i in range(0, self._radio._num_banks):
            index = type(self).bank_index[i]
            bank = self._radio._bank_class(self, index, "BANK-%s" % index)
            bank.index = i
            banks.append(bank)

        return banks


class ICT90_Alias(chirp_common.Alias):
    VENDOR = "Icom"
    MODEL = "IC-T90"


@directory.register
class ICx90Radio(icf.IcomCloneModeRadio):
    """Icom IC-E/T90"""
    VENDOR = "Icom"
    MODEL = "IC-E90"

    ALIASES = [ICT90_Alias]

    _model = "\x25\x07\x00\x01"
    _memsize = 0x2d40
    _endframe = "Icom Inc."

    _ranges = [(0x0000, 0x2d40, 32)]
    _num_banks = BANKS
    _bank_index_bounds = (0, BANK_NUM - 1)
    _can_hispeed = False

    def __init__(self, pipe):
        icf.IcomCloneModeRadio.__init__(self, pipe)

    def special_add(self, key, item_type, num, unique_idx):
        item = {}
        item["item_type"] = item_type
        item["num"] = num
        item["uidx"] = unique_idx
        self.special[key] = item

    def init_special(self):
        self.special = {}
        i = 0
        # program scan edges
        for x in range(25):
            self.special_add("Scan edge: %02dA" % x, "scan_edge", x * 2, i)
            self.special_add("Scan edge: %02dB" % x, "scan_edge", x * 2 + 1, i + 1)
            i += 2
        # call channels
        for x in range(5):
            self.special_add("Call ch: %d" % x, "call_chan", x, i)
            i += 1
        # VFO A
        for x in range(10):
            self.special_add("VFO A: %d" % x, "vfo_a", x, i)
            i += 1
        # VFO B
        for x in range(10):
            self.special_add("VFO B: %d" % x, "vfo_b", x, i)
            i += 1

    def get_sub_devices(self):
        return [ICx90Radio_ham(self._mmap), ICx90Radio_tv(self._mmap)]

    def clear_bank(self, loc):
        # it seems that empty invisible channel which isn't in bank is defined in bank_item by bytes 0x9f 0x00,
        # i.e. it has invisible_channel == 1
        # it seems that non-empty visible channel which isn't in bank is defined in bank_item by bytes 0x1f 0x00,
        # i.e. it has invisible_channel == 0
        # so do not touch the invisible_channel (the bit 7 of the first byte) here and only
        # set the rest of the bits
        self.memobj.banks[loc].bank_index = 0x1f
        self.memobj.banks[loc].prog_skip = 0
        self.memobj.banks[loc].mem_skip = 0
        self.memobj.banks[loc].bank_channel = 0

    # it seems the bank driver has different terminology about bank number and index
    # so in fact _get_bank and _set_bank are about indexes (i.e index in the array
    # of bank names - A .. Y
    # and _get_bank_index and _set_bank_index are about positions in the bank (0..99)
    def _get_bank(self, loc):
        i = self.memobj.banks[loc].bank_index
        return i if i < BANK_INDEX_NUM else None

    def _set_bank(self, loc, bank):
        if bank is None:
            self.clear_bank(loc)
        else:
            self.memobj.banks[loc].bank_index = bank
            # it seems if invisible_channel == 1 the channel is invisible (deleted)
            # so set it explicitly as visible
            self.memobj.banks[loc].invisible_channel = 0

    def _get_bank_index(self, loc):
        i = self.memobj.banks[loc].bank_channel
        return i if i < BANK_NUM else None

    def _set_bank_index(self, loc, index):
        self.memobj.banks[loc].bank_channel = index

    def get_features(self):
        self.init_special()
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_name = True
        rf.has_bank = True
        rf.has_bank_index = True
        rf.has_bank_names = False
        rf.can_delete = True
        rf.has_ctone = True
        rf.has_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_tuning_step = True
        rf.has_comment = False
        rf.memory_bounds = (0, MEM_NUM - 1)
        rf.has_sub_devices = True
        rf.valid_characters = CHARSET
        rf.valid_modes = list(MODE)
        rf.valid_tmodes = list(TONE_MODE)
        rf.valid_duplexes = list(DUPLEX)[:-1]
        rf.valid_tuning_steps = list(TUNE_STEP)
        rf.valid_bands = [(495000, 999990000)]
        rf.valid_skips = ["", "S", "P"]
        rf.valid_name_length = NAME_LENGTH
        rf.valid_special_chans = sorted(self.special.keys())

        return rf

    def map_dtmf_chirp2icom(self, item):
        item = str(item).upper()
        if item == "*":
            item = "E"
        elif item == "#":
            item = "F"
        elif item == " ":
            return 0xff
        try:
            ret = int(item, 16)
        except ValueError:
            raise errors.InvalidDataError("invalid DTMF number '%s'" % item)
        return ret

    def dtmf_chirp2icom(self, dtmf):
        return "".join(map(self.map_dtmf_chirp2icom, str(dtmf).rjust(DTMF_DIGITS_NUM)))

    def map_dtmf_icom2chirp(self, item):
        item = ord(item)
        if item == 0xff:
            return " "
        else:
            item &= 0x0f
            if item < 10:
                return str(item)
            else:
                return ["A", "B", "C", "D", "*", "#"][item - 10]

    def dtmf_icom2chirp(self, dtmf):
        return "".join(map(self.map_dtmf_icom2chirp, str(dtmf)))

    def apply_dtmf_autodial(self, setting, obj):
        obj = self.dtmf_chirp2icom(setting.value)

    def get_settings(self):
        try:
            _squelch = 1
            basic = RadioSettingGroup("basic", "Basic")
            expand_1 = RadioSettingGroup("expand_1", "Expand 1")
            expand_2 = RadioSettingGroup("expand_2", "Expand 2")
            dtmf_autodial = RadioSettingGroup("dtmf_autodial", "DTMF autodial")
            group = RadioSettings(basic, expand_1, expand_2, dtmf_autodial)

            # basic
            basic.append(RadioSetting("mem_channel", "Current memory channel",
                         RadioSettingValueInteger(0, MEM_NUM - 1, self.memobj.mem_channel)))
            basic.append(RadioSetting("squelch_level", "Squelch level",
                         RadioSettingValueList(SQUELCH_LEVEL,
                                               SQUELCH_LEVEL[self.memobj.squelch_level])))
            basic.append(RadioSetting("scan_resume", "Scan resume",
                         RadioSettingValueList(SCAN_RESUME,
                                               SCAN_RESUME[self.memobj.scan_resume])))
            basic.append(RadioSetting("scan_pause", "Scan pause",
                         RadioSettingValueList(SCAN_PAUSE,
                                               SCAN_PAUSE[self.memobj.scan_pause])))
            basic.append(RadioSetting("beep_volume", "Beep audio",
                         RadioSettingValueList(BEEP_VOLUME,
                                               BEEP_VOLUME[self.memobj.beep_volume])))
            basic.append(RadioSetting("beep", "Operation beep",
                         RadioSettingValueBoolean(self.memobj.beep)))
            basic.append(RadioSetting("backlight", "LCD backlight",
                         RadioSettingValueList(BACKLIGHT,
                                               BACKLIGHT[self.memobj.backlight])))
            basic.append(RadioSetting("busy_led", "Busy LED",
                         RadioSettingValueBoolean(self.memobj.busy_led)))
            basic.append(RadioSetting("auto_power_off", "Auto power off",
                         RadioSettingValueList(AUTO_POWER_OFF,
                                               AUTO_POWER_OFF[self.memobj.auto_power_off])))
            basic.append(RadioSetting("power_save", "Power save",
                         RadioSettingValueList(POWER_SAVE,
                                               POWER_SAVE[self.memobj.power_save])))
            basic.append(RadioSetting("monitor", "Monitor",
                         RadioSettingValueList(MONITOR,
                                               MONITOR[self.memobj.monitor])))
            basic.append(RadioSetting("dial_speedup", "Dial speedup",
                         RadioSettingValueBoolean(self.memobj.dial_speedup)))
            basic.append(RadioSetting("auto_repeater", "Auto repeater",
                         RadioSettingValueList(AUTO_REPEATER,
                                               AUTO_REPEATER[self.memobj.auto_repeater])))
            basic.append(RadioSetting("hm_75a_function", "HM-75A function",
                         RadioSettingValueList(HM_75A_FUNCTION,
                                               HM_75A_FUNCTION[self.memobj.hm_75a_function])))
            basic.append(RadioSetting("wx_alert", "WX alert",
                         RadioSettingValueBoolean(self.memobj.wx_alert)))
            basic.append(RadioSetting("wx_channel", "Current WX channel",
                         RadioSettingValueList(WX_CHANNEL,
                                               WX_CHANNEL[self.memobj.wx_channel])))
            basic.append(RadioSetting("comment", "Comment",
                         RadioSettingValueString(0, COMMENT_LEN,
                                                 str(self.memobj.comment),
                                                 autopad=True)))
            basic.append(RadioSetting("tune_step", "Current tune step",
                         RadioSettingValueList(TUNE_STEP_STR,
                                               TUNE_STEP_STR[self.memobj.tune_step])))
            basic.append(RadioSetting("band_selected", "Selected band",
                         RadioSettingValueInteger(0, BANDS - 1, self.memobj.band_selected)))
            basic.append(RadioSetting("memory_display", "Memory display",
                         RadioSettingValueList(MEMORY_DISPLAY,
                                               MEMORY_DISPLAY[self.memobj.memory_display])))
            basic.append(RadioSetting("memory_name", "Memory name",
                         RadioSettingValueBoolean(self.memobj.memory_name)))
            basic.append(RadioSetting("dial_select", "Dial select",
                         RadioSettingValueList(DIAL_SELECT,
                                               DIAL_SELECT[self.memobj.dial_select])))
            basic.append(RadioSetting("power", "RF power",
                         RadioSettingValueList(POWER,
                                               POWER[self.memobj.power])))
            basic.append(RadioSetting("vfo", "Current VFO",
                         RadioSettingValueList(VFO,
                                               VFO[self.memobj.vfo])))
            basic.append(RadioSetting("attenuator", "RF attenuator",
                         RadioSettingValueBoolean(self.memobj.attenuator)))
            basic.append(RadioSetting("skip_scan", "Skip scan",
                         RadioSettingValueBoolean(self.memobj.skip_scan)))
# TODO: this needs to be reverse engineered, because the following commented
# code does not seem correct
#            basic.append(RadioSetting("mode", "Current mode",
#                         RadioSettingValueList(OPERATION_MODE,
#                         OPERATION_MODE[self.memobj.mode])))
            basic.append(RadioSetting("vfo_scan", "VFO scan",
                         RadioSettingValueList(VFO_SCAN,
                                               VFO_SCAN[self.memobj.vfo_scan])))
            basic.append(RadioSetting("memory_scan", "Memory scan",
                         RadioSettingValueList(MEMORY_SCAN,
                                               MEMORY_SCAN[self.memobj.memory_scan])))
            basic.append(RadioSetting("tv_channel", "Current TV channel",
                         RadioSettingValueInteger(0, TV_CHANNELS - 1, self.memobj.tv_channel)))

            # DTMF auto dial
            dtmf_autodial.append(RadioSetting("autodial", "Autodial",
                                 RadioSettingValueList(AUTODIAL,
                                                       AUTODIAL[self.memobj.autodial])))
            dtmf_autodial.append(RadioSetting("dtmf_speed", "Speed",
                                 RadioSettingValueList(DTMF_SPEED,
                                                       DTMF_SPEED[self.memobj.dtmf_speed])))
            for x in range(DTMF_AUTODIAL_NUM):
                rs = RadioSetting("dtmf_codes[%d].dtmf_digits" % x, "DTMF autodial: %d" % x,
                                  RadioSettingValueString(0, DTMF_DIGITS_NUM,
                                                          self.dtmf_icom2chirp(self.memobj.dtmf_codes[x].dtmf_digits),
                                                          autopad=True, charset="0123456789ABCD*#abcd "))
                rs.set_apply_callback(self.apply_dtmf_autodial, self.memobj.dtmf_codes[x].dtmf_digits)
                dtmf_autodial.append(rs)

            # expand 1
            expand_1.append(RadioSetting("expand_1", "Expand 1",
                            RadioSettingValueBoolean(self.memobj.expand_1)))
            expand_1.append(RadioSetting("scan_stop_beep", "Scan stop beep",
                            RadioSettingValueBoolean(self.memobj.scan_stop_beep)))
            expand_1.append(RadioSetting("scan_stop_light", "Scan stop light",
                            RadioSettingValueBoolean(self.memobj.scan_stop_light)))
            expand_1.append(RadioSetting("light_postion", "Light position",
                            RadioSettingValueList(LIGHT_POSITION,
                                                  LIGHT_POSITION[self.memobj.light_position])))
            expand_1.append(RadioSetting("light_color", "Light color",
                            RadioSettingValueList(LIGHT_COLOR,
                                                  LIGHT_COLOR[self.memobj.light_color])))
            expand_1.append(RadioSetting("band_edge_beep", "Band edge beep",
                            RadioSettingValueBoolean(self.memobj.band_edge_beep)))
            expand_1.append(RadioSetting("auto_power_on", "Auto power on",
                            RadioSettingValueList(AUTO_POWER_ON,
                                                  AUTO_POWER_ON[self.memobj.auto_power_on])))
            expand_1.append(RadioSetting("key_lock", "Key lock",
                            RadioSettingValueList(KEY_LOCK,
                                                  KEY_LOCK[self.memobj.key_lock])))
            expand_1.append(RadioSetting("ptt_lock", "PTT lock",
                            RadioSettingValueBoolean(self.memobj.ptt_lock)))
            expand_1.append(RadioSetting("lcd_contrast", "LCD contrast",
                            RadioSettingValueList(LCD_CONTRAST,
                                                  LCD_CONTRAST[self.memobj.lcd_contrast])))
            expand_1.append(RadioSetting("opening_message", "Opening message",
                            RadioSettingValueBoolean(self.memobj.opening_message)))
            expand_1.append(RadioSetting("opening_message_text", "Opening message",
                            RadioSettingValueString(0, OPENING_MESSAGE_LEN,
                                                    str(self.memobj.opening_message_text),
                                                    autopad=True, charset=CHARSET)))

            # expand 2
            expand_2.append(RadioSetting("expand_2", "Expand 2",
                            RadioSettingValueBoolean(self.memobj.expand_2)))
            expand_2.append(RadioSetting("busy_lockout", "Busy lock out",
                            RadioSettingValueBoolean(self.memobj.busy_lockout)))
            expand_2.append(RadioSetting("timeout_timer", "Timeout timer",
                            RadioSettingValueList(TIMEOUT_TIMER,
                                                  TIMEOUT_TIMER[self.memobj.timeout_timer])))
            expand_2.append(RadioSetting("active_band", "Active band",
                            RadioSettingValueList(ACTIVE_BAND,
                                                  ACTIVE_BAND[self.memobj.active_band])))
            expand_2.append(RadioSetting("fm_narrow", "FM narrow",
                            RadioSettingValueBoolean(self.memobj.fm_narrow)))
            expand_2.append(RadioSetting("split", "Split",
                            RadioSettingValueBoolean(self.memobj.split)))
            expand_2.append(RadioSetting("morse_code_enable", "Morse code synthesizer",
                            RadioSettingValueBoolean(self.memobj.morse_code_enable)))
            expand_2.append(RadioSetting("morse_code_speed", "Morse code speed",
                            RadioSettingValueList(MORSE_CODE_SPEED,
                                                  MORSE_CODE_SPEED[self.memobj.morse_code_speed])))

            return group
        except:
            import traceback
            LOG.error("Failed to parse settings: %s", traceback.format_exc())
            return None

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue
            try:
                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    try:
                        element.run_apply_callback()
                    except NotImplementedError as e:
                        LOG.error("icx90: %s", e)
                    continue

                # Find the object containing setting.
                obj = self.memobj
                bits = element.get_name().split(".")
                setting = bits[-1]
                for name in bits[:-1]:
                    if name.endswith("]"):
                        name, index = name.split("[")
                        index = int(index[:-1])
                        obj = getattr(obj, name)[index]
                    else:
                        obj = getattr(obj, name)

                try:
                    old_val = getattr(obj, setting)
                    LOG.debug("Setting %s(%r) <= %s" % (
                        element.get_name(), old_val, element.value))
                    setattr(obj, setting, element.value)
                except AttributeError as e:
                    LOG.error("Setting %s is not in the memory map: %s" %
                              (element.get_name(), e))
            except Exception:
                LOG.debug(element.get_name())
                raise

    def process_mmap(self):
        self.memobj = bitwise.parse(ICX90_MEM_FORMAT, self._mmap)

    def sync_in(self):
        icf.IcomCloneModeRadio.sync_in(self)
        self.process_mmap()

    def sync_out(self):
        icf.IcomCloneModeRadio.sync_out(self)

    def freq_chirp2icom(self, freq):
        if chirp_common.is_fractional_step(freq):
            mult = 6250
            multr = 1
        else:
            mult = 5000
            multr = 0

        return (freq / mult, multr)

    def freq_icom2chirp(self, freq, mult):
        return freq * (6250 if mult else 5000)

    def get_skip(self, number):
        bank_item = self.memobj.banks[number]
        if bank_item.prog_skip:
            return "P"
        elif bank_item.mem_skip:
            return "S"
        return ""

    def set_skip(self, number, skip):
        bank_item = self.memobj.banks[number]
        if skip == "P":
            bank_item.prog_skip = 1
            bank_item.mem_skip = 1
        elif skip == "S":
            bank_item.prog_skip = 0
            bank_item.mem_skip = 1
        elif skip == "":
            bank_item.prog_skip = 0
            bank_item.mem_skip = 0
        else:
            raise errors.InvalidDataError("skip '%s' not supported" % skip)

    # returns (memobj.mem_item, is_special_channel, unique_idx)
    def get_mem_item(self, number):
        try:
            item_type = self.special[number]["item_type"]
            num = self.special[number]["num"]
            unique_idx = self.special[number]["uidx"]
            if item_type == "scan_edge":
                return (self.memobj.scan_edges[num], True, unique_idx)
            elif item_type == "call_chan":
                return (self.memobj.call_channels[num], True, unique_idx)
            elif item_type == "vfo_a":
                return (self.memobj.vfo_a_band[num], True, unique_idx)
            elif item_type == "vfo_b":
                return (self.memobj.vfo_b_band[num], True, unique_idx)
            else:
                raise errors.InvalidDataError("unknown special channel type '%s'" % item_type)
        except KeyError:
            return (self.memobj.memory[number], False, number)

    def get_raw_memory(self, number):
        (mem_item, special, unique_idx) = self.get_mem_item(number)
        return repr(mem_item)

    def get_memory(self, number):
        mem = chirp_common.Memory()

        (mem_item, special, unique_idx) = self.get_mem_item(number)

        freq = self.freq_icom2chirp(mem_item.freq, mem_item.freq_mult)
        if freq == 0:
            mem.empty = True
        else:
            mem.empty = False
            mem.freq = freq
            if special:
                mem.name = " " * NAME_LENGTH
            else:
                mem.name = str(mem_item.name).rstrip("\x00 ")
            mem.rtone = chirp_common.TONES[(mem_item.tx_tone_hi << 4) + mem_item.tx_tone_lo]
            mem.ctone = chirp_common.TONES[mem_item.rx_tone]
            mem.dtcs = chirp_common.DTCS_CODES[mem_item.dtcs]
            mem.dtcs_polarity = DTCS_POLARITY[mem_item.dtcs_polarity]
            mem.offset = self.freq_icom2chirp(mem_item.offset_freq, mem_item.offset_freq_mult)
            mem.duplex = DUPLEX[mem_item.duplex]
            mem.tmode = TONE_MODE[mem_item.tone_mode]
            mem.tuning_step = TUNE_STEP[mem_item.tune_step]
            mem.mode = MODE[mem_item.mode]
            if not special:
                mem.skip = self.get_skip(number)
        if special:
            mem.extd_number = number
            mem.number = -len(self.special) + unique_idx
        else:
            mem.number = number

        return mem

    def set_memory(self, memory):
        (mem_item, special, unique_idx) = self.get_mem_item(
            memory.extd_number or memory.number)
        if memory.empty:
            mem_item.set_raw("\x00" * MEM_ITEM_SIZE)
            self.clear_bank(memory.number)
            self.memobj.banks[memory.number].invisible_channel = 1
        else:
            (mem_item.freq, mem_item.freq_mult) = self.freq_chirp2icom(memory.freq)
            if special:
                mem_item.name = " " * NAME_LENGTH
            else:
                self.memobj.banks[memory.number].invisible_channel = 0

                for x in range(NAME_LENGTH):
                    try:
                        mem_item.name[x] = str(memory.name[x])
                    except IndexError:
                        mem_item.name[x] = " "
            mem_item.tx_tone_hi = chirp_common.TONES.index(memory.rtone) >> 4
            mem_item.tx_tone_lo = chirp_common.TONES.index(memory.rtone) & 0x0f
            mem_item.rx_tone = chirp_common.TONES.index(memory.ctone)
            mem_item.dtcs = chirp_common.DTCS_CODES.index(memory.dtcs)
            mem_item.dtcs_polarity = DTCS_POLARITY.index(memory.dtcs_polarity)
            (mem_item.offset_freq, mem_item.offset_freq_mult) = self.freq_chirp2icom(memory.offset)
            mem_item.duplex = DUPLEX.index(memory.duplex)
            mem_item.tone_mode = TONE_MODE.index(memory.tmode)
            mem_item.tune_step = TUNE_STEP.index(memory.tuning_step)
            mem_item.mode = MODE.index(memory.mode)
            if not special:
                self.set_skip(memory.number, memory.skip)

    def get_bank_model(self):
        return ICx90BankModel(self)


class ICx90Radio_ham(ICx90Radio):
    VARIANT = 'Radio'

    def get_features(self):
        rf = ICx90Radio.get_features(self)
        rf.has_sub_devices = False

        return rf


class ICx90Radio_tv(ICx90Radio):
    VARIANT = "TV"

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_name = True
        rf.has_bank = False
        rf.has_bank_index = False
        rf.has_bank_names = False
        rf.can_delete = True
        rf.has_ctone = False
        rf.has_dtcs = False
        rf.has_dtcs_polarity = False
        rf.has_tuning_step = False
        rf.has_comment = False
        rf.has_sub_devices = False
        rf.memory_bounds = (0, TV_CHANNELS - 1)
        rf.valid_characters = CHARSET
        rf.valid_modes = list(TV_MODE)
        rf.valid_tmodes = []
        rf.valid_duplexes = [""]
        rf.valid_tuning_steps = []
        rf.valid_bands = [(46750000, 957750000)]
        rf.valid_skips = ["", "S"]
        rf.valid_name_length = TV_NAME_LENGTH
        rf.valid_special_chans = []

        return rf

    def get_bank_model(self):
        return None

    def freq_chirp2icom(self, freq):
        return freq / 5000

    def freq_icom2chirp(self, freq):
        return freq * 5000

    def get_skip(self, number):
        if self.memobj.tv_channel_skip[number] == 2:
            return "S"
        return ""

    def set_skip(self, number, skip):
        if skip == "S":
            self.memobj.tv_channel_skip[number] = 2
        elif skip == "":
            self.memobj.tv_channel_skip[number] = 0
        else:
            raise errors.InvalidDataError("skip '%s' not supported" % skip)

    def get_raw_memory(self, number):
        return repr(self.memobj.tv_memory[number])

    def get_memory(self, number):
        mem = chirp_common.Memory()

        mem_item = self.memobj.tv_memory[number]

        mem.freq = self.freq_icom2chirp(mem_item.freq)
        if self.memobj.tv_channel_skip[number] == 1:
            mem.empty = True
            mem.mode = TV_MODE[0]
        else:
            mem.empty = False
            mem.name = str(mem_item.name).rstrip("\x00 ")
            mem.mode = TV_MODE[mem_item.mode]
            mem.skip = self.get_skip(number)
        mem.number = number

        return mem

    def set_memory(self, memory):
        mem_item = self.memobj.tv_memory[memory.number]
        if memory.empty:
            self.memobj.tv_channel_skip[memory.number] = 1
            mem_item.set_raw("\x00" * TV_MEM_ITEM_SIZE)
        else:
            mem_item.freq = self.freq_chirp2icom(memory.freq)
            for x in range(TV_NAME_LENGTH):
                try:
                    mem_item.name[x] = str(memory.name[x])
                except IndexError:
                    mem_item.name[x] = " "

            mem_item.mode = TV_MODE.index(memory.mode)
            self.set_skip(memory.number, memory.skip)


def dump_banks(icx90, template_file):
    mb = icx90.get_features().memory_bounds
    with open(template_file, "w") as f:
        for mi in range(mb[0], mb[1] + 1):
            mem = icx90.get_memory(mi)
            bank = icx90._get_bank(mi)
            if bank is not None:
                bank = BANK_INDEX[bank]
            bank_pos = icx90._get_bank_index(mi)
            if not mem.empty and bank is not None and bank_pos is not None:
                f.write("%s;%s;%d\n" % (mem.name, bank, bank_pos))


def read_template_file(template_file):
    banks_templ = {}
    with open(template_file, "r") as f:
        for line in f:
            l = line.split(";")
            l[1] = BANK_INDEX.index(l[1])
            banks_templ[l[0]] = l[1:]
    return banks_templ


def reorder_banks(icx90, preserve_position, preserve_unknown, banks_templ):
    banks_cnt = []
    mb = icx90.get_features().memory_bounds
    for i in range(0, BANKS):
        banks_cnt.append(0)
    for mi in range(mb[0], mb[1] + 1):
        mem = icx90.get_memory(mi)
        if preserve_unknown:
            bank = icx90._get_bank(mi)
            bank_pos = icx90._get_bank_index(mi)
        else:
            bank = None
            bank_pos = 0
        if not mem.empty:
            if mem.name in banks_templ:
                bank = int(banks_templ[mem.name][0])
                if preserve_position:
                    bank_pos = int(banks_templ[mem.name][1])
                else:
                    bank_pos = banks_cnt[bank]
                    banks_cnt[bank] += 1
                print("%s\t-> %s, %02d" % (mem.name, BANK_INDEX[bank], bank_pos))
            # explicitly set non empty channel as visible
            icx90.memobj.banks[mi].invisible_channel = 0
        if bank >= BANKS or bank_pos >= BANK_NUM:
            bank = None
            bank_pos = 0
        icx90._set_bank(mi, bank)
        icx90._set_bank_index(mi, bank_pos)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Icom IC-E90 banks handling helper.")
    parser.add_argument("icx90_img_file", help="IC-X90 IMG file.")
    parser.add_argument("template_file", help="Banks template file.")
    parser.add_argument("-r", "--read-banks", action="store_true",
                        help="Read banks content and store it to template file.")
    parser.add_argument("-f", "--fix-banks", action="store_true",
                        help="Fix banks content, reorder channels in banks according to the provided template file and write it to IMG file.")
    parser.add_argument("-p", "--preserve-position", action="store_true",
                        help="Preserve bank position as in the template file if possible, otherwise put channels to banks according their position in the memory.")
    parser.add_argument("-u", "--preserve-unknown", action="store_true",
                        help="Preserve channels in banks if they aren't in the template file and don't conflicts with the template file, otherwise remove them from the banks.")
    parser.add_argument("-b", "--backup_img", action="store_true",
                        help="Backup IMG file before changing it.")

    args = vars(parser.parse_args())

    options_ok = True
    try:
        icx90_img_file = str(args.pop("icx90_img_file"))
        template_file = str(args.pop("template_file"))
    except KeyError:
        options_ok = False

    if options_ok and (args["read_banks"] or args["fix_banks"]):
        icx90 = ICx90Radio(icx90_img_file)
        if args["read_banks"]:
            dump_banks(icx90, template_file)
        elif args["fix_banks"]:
            banks_templ = read_template_file(template_file)
            if args["backup_img"]:
                icx90.save_mmap(icx90_img_file + ".bak")
            reorder_banks(icx90, args["preserve_position"], args["preserve_unknown"], banks_templ)
            icx90.save_mmap(icx90_img_file)
    else:
        parser.print_help()
